import datetime
import logging
import os
import sys
import termios
import time
import traceback
from queue import LifoQueue
from threading import Thread

import readchar
from rich.columns import Columns
from rich.panel import Panel
from rich.text import Text
from rich.live import Live
from rich.table import Table
from rich.layout import Layout
from rich.align import Align
from rich import box


from dry_pipe.monitoring import PipelineMetricsTable, fetch_task_group_metrics
from dry_pipe.pipeline import PipelineInstance, Pipeline
from dry_pipe.script_lib import ps_resources
from dry_pipe.task_state import TaskState

logger = logging.getLogger(__name__)


class CliScreen:

    def __init__(self, pipeline_instance_dir, regex_grouper, is_monitor_mode, quit_listener=None, group_by=None):
        self.is_monitor_mode = is_monitor_mode
        self.pipeline_instance_dir = pipeline_instance_dir
        self.pipeline_hints = PipelineInstance.load_hints(pipeline_instance_dir)
        self.quit = False
        self.screen = Summary(self)
        self.selected_failed_task = None
        self.prompt = None
        self.error_msg = []
        self.loop_counter = 0
        self.rich_live_auto_refresh = True
        self.quit_listener = quit_listener

        stdin_fd = sys.stdin.fileno()
        old_tty_settings = termios.tcgetattr(stdin_fd)

        self.cleanup_tty = lambda:\
            termios.tcsetattr(stdin_fd, termios.TCSADRAIN, old_tty_settings)

        self.queue = LifoQueue(maxsize=2)

        pipeline = Pipeline.load_from_module_func(self.pipeline_hints["pipeline"])
        self.task_groupers = pipeline.task_groupers

        if regex_grouper is not None:
            self.task_groupers["regex_grouper"] = regex_grouper
            self.selected_grouper_name = "regex_grouper"
        else:
            self.selected_grouper_name = "by_type"

        if os.environ.get("DRYPIPE_CLI_AUTO_REFRESH") == "False":
            self.rich_live_auto_refresh = False

        self.listen_keyboard_enabled = True

        if os.environ.get("DRYPIPE_CLI_LISTEN_KEYBOARD_ENABLED") == "False":
            self.listen_keyboard_enabled = False

        if group_by is not None:

            g = pipeline.task_groupers.get(group_by)
            if g is None:
                raise Exception(
                    f"pipeline was not instantiated with task_groupers dict containing {group_by}"
                )
            self.selected_grouper_name = group_by

        class ShallowPipelineInstance:
            def __init__(self):
                self.pipeline_instance_dir = pipeline_instance_dir

        self.pipeline_instance = ShallowPipelineInstance()

    def request_quit(self):
        self.quit = True

    def press(self, key):
        logger.debug("key pressed: %s", key)
        try:
            if key == "q" or key == readchar.key.CTRL_C:
                self.request_quit()
                self.queue.put("m")
            elif key == readchar.key.UP:
                logger.debug("UP")
                self.queue.put("up")
            elif key == readchar.key.DOWN:
                self.queue.put("down")
                logger.debug("DOWN")
            elif key == readchar.key.ENTER:
                self.queue.put("enter")
            elif key == readchar.key.LEFT:
                self.queue.put("left")
            elif key == "f":
                self.queue.put("failed-screen")
            elif key == "s":
                self.queue.put("summary-screen")
            elif key == "r":
                self.queue.put("restart")
            elif key == "y":
                self.queue.put("y")
            elif key == "n":
                self.queue.put("n")
        except Exception:
            logger.exception(f"failed on key command {key}")
            self.error_msg.append(traceback.format_exc())

    def is_prompt_restart_task(self):
        return self.prompt == "restart-task"

    def set_prompt_restart_task(self):
        self.prompt = "restart-task"

    def cancel_prompt(self):
        self.prompt = None

    def start(self):
        def refresher():
            refresh_per_second = 4
            if self.rich_live_auto_refresh:
                refresh_per_second = 1
            with Live(auto_refresh=self.rich_live_auto_refresh, refresh_per_second=refresh_per_second) as live:
                try:
                    while not self.quit:
                        msg = self.queue.get()
                        if msg == "failed-screen":
                            self.screen = FailedTaskList(self)
                            self.screen.reload()
                        elif msg == "summary-screen":
                            self.screen = Summary(self)
                            self.screen.reload()

                        self.screen.refresh(live, self._main_screen(), msg)

                except Exception:
                    self.request_quit()
                    logger.exception(f"failed in refresh loop")
                    self.error_msg.append(traceback.format_exc())

            logger.info(f"out of Live loop")

            self.quit_listener(self)

        logger.info("will start screen refresher thread")
        Thread(target=refresher).start()

        def data_loader_thread():
            while not self.quit:
                try:
                    self.screen.reload()
                    self.queue.put("reload")
                    time.sleep(3)
                except Exception as ex:
                    self.request_quit()
                    logger.exception(ex)
                    break

        Thread(target=data_loader_thread).start()

        if self.listen_keyboard_enabled:

            readchar.config.INTERRUPT_KEYS = []

            def _listen_keyboard():
                logger.info("_listen_keyboard thread started")
                while True:
                    try:
                        c = readchar.readkey()
                        self.press(c)

                        if self.quit:
                            break
                        if c == "q":
                            break
                    except Exception as ex:
                        self.cleanup_tty()
                        self.request_quit()
                        logger.exception(ex)
                        break

                logger.debug("out of _listen_keyboard loop")

            listen_keyboard_thread = Thread(target=_listen_keyboard)
            listen_keyboard_thread.start()
        else:
            logger.info("will NOT listen_keyboard")


    def _main_screen(self):
        layout = Layout()
        layout.split_column(
            Layout(name="header1", size=1),
            Layout(name="header2", size=1),
            Layout(name="header3", size=1),
            Layout(name="header4", size=1),
            Layout(name="body", ratio=10),
            Layout(name="footer", minimum_size=1)
        )

        layout["footer"].update("+++")

        pipeline_mod_func = self.pipeline_hints.get("pipeline")
        if self.is_monitor_mode:
            console_mode = f"  [bright_yellow]console in monitor mode (no launching)[/]  ({self.loop_counter})"
        else:
            console_mode = f"  [bright_yellow]console in launcher mode[/]  ({self.loop_counter})"

        if pipeline_mod_func is not None:
            title = f"[bold]DryPipe([green1]{pipeline_mod_func}[/], "
            title += f"[cyan1]{self.pipeline_instance_dir}[/])  {console_mode}"
        else:
            title = f"DryPipe - {console_mode}"

        help_panel = Columns([
            "command keys: "
            "[b]s[/]: summary",
            "[b]f[/]: failed tasks",
            "[b]q[/]: quit"
        ], equal=True, expand=False)

        layout["header1"].update(Align(title, align="center"))
        layout["header2"].update(Align(help_panel, align="left"))

        return layout


def styled_prompt(question, choices, default):
    return f"[cyan1]{question}[/] [purple]{','.join(choices)}[/] ({default}):"


class TaskView:

    def __init__(self, prev_screen, task_key):
        self.cli_screen = prev_screen.cli_screen
        self.prev_screen = prev_screen
        self.task_key = task_key
        self.task_state = TaskState.from_task_control_dir(
            self.cli_screen.pipeline_instance_dir, self.task_key
        )
        self.task_resources_view = TaskResourcesView(self.task_state)
        self.prompt_restart = False
        self.restart_in_progress = False

    def refresh(self, live, layout, msg=None):
        if msg == "left":
            self.cli_screen.screen = self.prev_screen
            self.cli_screen.screen.reload()
            self.cli_screen.screen.refresh(live, layout)
        else:
            if self.prompt_restart:
                if msg == "y":
                    self.prompt_restart = False
                    self.restart_in_progress = True
                    self.task_state.transition_to_queued()
                elif msg == "n":
                    self.prompt_restart = False

            if msg == "restart":
                self.prompt_restart = True


            out, err, log = self.data

            def g():
                for step_number, started, completed_or_none in self.task_state.history_steps_time_intervals():
                    if completed_or_none is None:
                        yield f"step-{step_number}(started at {started.strftime('%Hh:%M:%S%z')}, ...)"
                    else:
                        dt = (completed_or_none - started)
                        dt = datetime.timedelta(seconds=round(dt.total_seconds(), 0) or 1)
                        yield f"step-{step_number}({dt})"

            history_line = " -> ".join(g())

            def task_attr(v):
                return f"[cyan1]{v}[/]"

            task_title = f"Task(key={task_attr(self.task_key)}, state={task_attr(self.task_state.state_name)}"

            if self.task_resources_view.pid is not None:
                task_title += f", pid={task_attr(self.task_resources_view.pid)})"
            elif self.task_resources_view.slurm_job_id is not None:
                task_title += f", slurm_job_id={task_attr(self.task_resources_view.slurm_job_id)})"
            else:
                task_title += ")"

            layout["header3"].update(Align(task_title, align="center"))

            if self.prompt_restart:
                layout["header4"].update(Align(styled_prompt("restart task ?", ["y", "n"], "y"), align="center"))
            elif self.restart_in_progress:
                layout["header4"].update(Align("...restarting"))
                self.restart_in_progress = False
            else:
                layout["header4"].update(Text(f"History[{history_line}]"))

            task_ctrl_layout = Layout()
            task_ctrl_layout.split_column(
                Text("r: restart"),
                Text("k: kill (SIGTERM)"),
            )

            layout3 = Layout()
            layout3.split_column(
                self.task_resources_view.render(),
                Panel(Text(log, style="blue"), title="drypipe janitor log(tail -f drypipe.log)")
            )
            main_layout = Layout()
            main_layout.split_row(
                Panel(Text(out, style="green"), title="tail -f out.log"),
                Panel(Text(err, style="red"), title="tail -f err.log"),
                layout3
            )
            layout["body"].update(main_layout)
            live.update(Panel(layout, border_style="blue"), refresh=True)

    def reload(self):
        self.task_state = TaskState.from_task_control_dir(
            self.cli_screen.pipeline_instance_dir, self.task_key
        )
        out, err, log = self.task_state.tail_out_err_drypipe()
        self.data = [out, err, log]
        self.task_resources_view.reload()


class TaskResourcesView:

    def __init__(self, task_state):
        self.task_state = task_state
        self.pid = None
        self.slurm_job_id = None
        self.task_ps_resources = None

    def reload(self):
        self.pid = self.task_state.pid()
        if self.pid is not None:
            r = list(ps_resources(self.pid))
            if len(r) == 0:
                self.task_ps_resources = None
            else:
                self.task_ps_resources = r

        else:
            self.slurm_job_id = self.task_state.slurm_job_id()
            if self.slurm_job_id is not None:
                self.task_ps_resources = [["na"], ["na"]]
            else:
                self.task_ps_resources = None

    def render(self):
        if self.task_ps_resources is None:
            return Panel(Text("Task Not Running"))
        else:
            resource_headers, resource_values = self.task_ps_resources
            resource_table = Table(
                show_header=True,
                header_style="none",
                expand=True,
                box=box.ROUNDED
            )
            for h in resource_headers:
                resource_table.add_column(h, style="bold")
            resource_table.add_row(*[str(v) for v in resource_values], style="green1")
            return resource_table


class FailedTaskList:

    def __init__(self, cli_screen):
        self.selected_row = None
        self.cli_screen = cli_screen
        self.selected_task_key = None
        self.failed_task_states = []

    def refresh(self, live, layout, msg=None):

        if msg in ["up", "down"]:
            if self.selected_row is None:
                self.selected_row = 0
            else:
                if msg == "up":
                    if self.selected_row > 0:
                        self.selected_row -= 1
                else:
                    row_count = len(self.failed_task_states)
                    if self.selected_row < (row_count - 1):
                        self.selected_row += 1
        elif msg == "enter":
            if self.selected_row is None:
                self.selected_row = 0
            self.cli_screen.screen = TaskView(self, self.failed_task_states[self.selected_row].task_key)
            self.cli_screen.screen.reload()
            self.cli_screen.screen.refresh(live, layout)
            return
        elif msg == "restart":
            #TODO
            pass

        elif msg == "left":
            self.cli_screen.press("s")

        layout["header3"].update("Failed Tasks")

        fail_count = len(self.failed_task_states)
        if fail_count == 0:
            layout["header4"].update(Text(f"there are no failed tasks"))
        else:
            layout["header4"].update("")

        table = Table(
            show_header=False,
            expand=True,
            show_lines=True
        )
        table.add_column(ratio=1)
        table.add_column(ratio=9)
        c = 0
        for task_state in self.failed_task_states:
            err = task_state.tail_err_if_failed(4)
            if self.selected_row == c:
                selected_prefix = " -> "
                selected_style = "bold"
            else:
                selected_prefix = "    "
                selected_style = ""

            table.add_row(
                Text(f"{selected_prefix} {task_state.task_key}"),
                Text(err, style=f"red"),
                style=selected_style
            )
            c += 1

        layout["body"].update(table)
        live.update(Panel(layout, border_style="blue"), refresh=True)

    def reload(self):
        self.failed_task_states = sorted(
            [
                task_state
                for task_state in TaskState.failed_task_states(self.cli_screen.pipeline_instance_dir)
            ],
            key=lambda s: s.task_key
        )


class Summary:

    def __init__(self, cli_screen):
        self.selected_row = None
        self.cli_screen = cli_screen
        self.data = None
        self.group_selected = None
        self.selected_task_key = None

    def refresh(self, live, layout, msg=None):
        if self.data is None:
            time.sleep(1)
        else:

            header, body, footer = self.data

            if msg in ["up", "down"]:
                if self.selected_row is None:
                    self.selected_row = 0
                else:
                    if msg == "up":
                        if self.selected_row > 0:
                            self.selected_row -= 1
                    else:
                        row_count = len(body)
                        if self.selected_row < (row_count - 1):
                            self.selected_row += 1
            elif msg == "enter":
                if self.group_selected is None:
                    if self.selected_row is not None:
                        self.group_selected = body[self.selected_row][0]
                        self.selected_row = None
                        logger.debug(f"group_selected %s", self.group_selected)
                        header, body, footer = self.reload()
                else:
                    if self.selected_row is None:
                        self.selected_row = 0
                    self.cli_screen.screen = TaskView(self, body[self.selected_row][0])
                    self.cli_screen.screen.reload()
                    self.cli_screen.screen.refresh(live, layout)
                    return

            elif msg == "left" and self.group_selected is not None:
                self.group_selected = None
                logger.debug(f"group_selected %s", self.group_selected)
                header, body, footer = self.reload()

            table = Table(
                show_header=True,
                header_style="none",
                show_lines=False,
                show_edge=False,
                expand=True
            )

            for h in header:
                table.add_column(h)

            def color_row(row, is_selected=False):
                c = 0
                row_cells = []
                for v in row:
                    selected_style = "bold" if is_selected else ""
                    if v is None or v == 0:
                        row_cells.append("")
                    else:
                        if c == 4:
                            row_cells.append(f"[green1 {selected_style}]{v}[/]")
                        elif c == 5:
                            row_cells.append(f"[blue {selected_style}]{v}[/]")
                        elif c in [6, 7, 8]:
                            row_cells.append(f"[red {selected_style}]{v}[/]")
                        elif c == 0:

                            if self.group_selected is None:
                                parent_group = ""
                            else:
                                parent_group = f"{self.group_selected}/"
                            selected_cursor = "-> " if is_selected else "   "
                            row_cells.append(f"[yellow {selected_style}]{selected_cursor}{parent_group}{v}[/]")
                        elif c == 1:
                            row_cells.append(f"[yellow {selected_style}]{v}[/]")
                        else:
                            row_cells.append(f"[yellow {selected_style}]{v}[/]")
                    c += 1
                return row_cells

            r = 0
            for row in body:
                is_selected = self.selected_row is not None and self.selected_row == r
                table.add_row(*color_row(row, is_selected))
                r += 1

            table.add_row(*color_row(footer))

            layout["header3"].update("Task Execution Status Summary")
            layout["header4"].update("")
            layout["body"].update(table)
            live.update(Panel(layout, border_style="blue"), refresh=True)

    def reload(self):
        task_grouper = self.cli_screen.task_groupers[self.cli_screen.selected_grouper_name]

        if self.group_selected is None:
            header, body, footer = PipelineMetricsTable.summarized_table_from_task_group_metrics(
                fetch_task_group_metrics(
                    self.cli_screen.pipeline_instance.pipeline_instance_dir,
                    task_grouper
                )
            )
        else:
            def task_filter(task_key):
                task_group = task_grouper(task_key)
                return task_group == self.group_selected

            header, body, footer = PipelineMetricsTable.summarized_table_from_task_group_metrics(
                fetch_task_group_metrics(
                    self.cli_screen.pipeline_instance.pipeline_instance_dir,
                    task_key_grouper=lambda i: i,
                    task_filter=task_filter
                )
            )

        self.data = [header, body, footer]
        return header, body, footer





import unittest

from dry_pipe.pubsub import subscription_reducer, subscription_registry_reducer


class WebSocketPubSubTests(unittest.TestCase):

    def test_subscription_reducer(self):

        state = {}

        state = subscription_reducer(state, {
            "name": "init",
            "sid": "s1"
        })

        self.assertDictEqual({
            "sid": "s1"
        }, state)

        state = subscription_reducer(state, {
            "name": "observePipeline",
            "pid": "pA",
            "sid": "s1"
        })

        state = subscription_reducer(state, {
            "name": "observeTask",
            "task_key": "tA",
            "sid": "s1"
        })

        self.assertDictEqual({
            "sid": "s1",
            "pid": "pA",
            "task_key": "tA"
        }, state)

        state = subscription_reducer(state, {
            "name": "unObserveTask"
        })

        self.assertDictEqual({
            "sid": "s1",
            "pid": "pA"
        }, state)

    def test_subscription_registry_reducer_01(self):

        base_reg = {
            "sids_to_subscriptions": {},
            "pids_to_sids": {},
            "task_keys_to_sids": {}
        }

        sub0, reg0 = subscription_registry_reducer(None, None)

        self.assertDictEqual(base_reg, reg0)

        sub1, reg1 = subscription_registry_reducer(reg0, {
            "name": "init",
            "sid": "s1"
        })

        self.assertDictEqual({
            "sids_to_subscriptions": {
                "s1": {
                    "sid": "s1"
                }
            },
            "pids_to_sids": {},
            "task_keys_to_sids": {}
        }, reg1)

        sub2, reg2 = subscription_registry_reducer(reg1, {
            "name": "observePipeline",
            "pid": "pA",
            "sid": "s1"
        })

        self.assertDictEqual({
            "sids_to_subscriptions": {
                "s1": {
                    "sid": "s1",
                    "pid": "pA"
                }
            },
            "pids_to_sids": {
                "pA": ["s1"]
            },
            "task_keys_to_sids": {}
        }, reg2)

        sub3, reg3 = subscription_registry_reducer(reg2, {
            "name": "init",
            "sid": "s2"
        })

        self.assertDictEqual({
            "sids_to_subscriptions": {
                "s1": {
                    "sid": "s1",
                    "pid": "pA"
                },
                "s2": {
                    "sid": "s2"
                }
            },
            "pids_to_sids": {
                "pA": ["s1"]
            },
            "task_keys_to_sids": {}
        }, reg3)

        sub4, reg4 = subscription_registry_reducer(reg3, {
            "name": "observePipeline",
            "pid": "pA",
            "sid": "s2"
        })

        self.assertDictEqual({
            "sids_to_subscriptions": {
                "s1": {
                    "sid": "s1",
                    "pid": "pA"
                },
                "s2": {
                    "sid": "s2",
                    "pid": "pA"
                }
            },
            "pids_to_sids": {
                "pA": ["s1", "s2"]
            },
            "task_keys_to_sids": {}
        }, reg4)

        sub5, reg5 = subscription_registry_reducer(reg4, {
            "name": "observeTask",
            "task_key": "tA",
            "sid": "s2"
        })

        self.assertDictEqual({
            "sids_to_subscriptions": {
                "s1": {
                    "sid": "s1",
                    "pid": "pA"
                },
                "s2": {
                    "sid": "s2",
                    "pid": "pA",
                    "task_key": "tA"
                }
            },
            "pids_to_sids": {
                "pA": ["s1", "s2"]
            },
            "task_keys_to_sids": {
                "tA": ["s2"]
            }
        }, reg5)

        sub6, reg6 = subscription_registry_reducer(reg5, {
            "name": "observeTask",
            "task_key": "tB",
            "sid": "s2"
        })

        self.assertDictEqual({
            "sids_to_subscriptions": {
                "s1": {
                    "sid": "s1",
                    "pid": "pA"
                },
                "s2": {
                    "sid": "s2",
                    "pid": "pA",
                    "task_key": "tB"
                }
            },
            "pids_to_sids": {
                "pA": ["s1", "s2"]
            },
            "task_keys_to_sids": {
                "tB": ["s2"]
            }
        }, reg6)

        sub7, reg7 = subscription_registry_reducer(reg6, {
            "name": "unObservePipeline",
            "sid": "s2"
        })

        self.assertDictEqual({
            "sids_to_subscriptions": {
                "s1": {
                    "sid": "s1",
                    "pid": "pA"
                },
                "s2": {
                    "sid": "s2"
                }
            },
            "pids_to_sids": {
                "pA": ["s1"]
            },
            "task_keys_to_sids": {}
        }, reg7)

        sub8, reg8 = subscription_registry_reducer(reg7, {
            "name": "close",
            "sid": "s1"
        })

        self.assertDictEqual({
            "sids_to_subscriptions": {
                "s2": {
                    "sid": "s2"
                }
            },
            "pids_to_sids": {},
            "task_keys_to_sids": {}
        }, reg8)

        return base_reg, reg1, reg2, reg3, reg4, reg5, reg6, reg7, reg8

    def test_subscription_registry_reducer_02(self):

        base_reg, reg1, reg2, reg3, reg4, reg5, reg6, reg7, reg8 = self.test_subscription_registry_reducer_01()

        subs, reg = subscription_registry_reducer(reg5, {
            "name": "close_sessions",
            "sids": ["s1", "s2"]
        })

        self.assertDictEqual(base_reg, reg)

        subs, reg = subscription_registry_reducer(reg5, {
            "name": "close_sessions",
            "sids": ["s1"]
        })

        self.assertDictEqual({
            "sids_to_subscriptions": {
                "s2": {
                    "sid": "s2",
                    "pid": "pA",
                    "task_key": "tA"
                }
            },
            "pids_to_sids": {
                "pA": ["s2"]
            },
            "task_keys_to_sids": {
                "tA": ["s2"]
            }
        }, reg)

        for reg in [base_reg, reg1, reg2, reg3, reg4, reg5, reg6, reg7, reg8]:

            sids_to_subscriptions = reg["sids_to_subscriptions"]

            subs, reg_after = subscription_registry_reducer(reg, {
                "name": "reset_from_sids_to_subscriptions",
                "sids_to_subscriptions": sids_to_subscriptions
            })

            self.assertDictEqual(reg, reg_after)



import collections
import inspect
import json
import os
import re
import sys
import textwrap
from fnmatch import fnmatch

from dry_pipe.script_lib import PortablePopen, parse_ssh_specs
from dry_pipe.utils import bash_shebang
from dry_pipe.internals import \
    Executor, Local, PreExistingFile, IndeterminateFile, ProducedFile, \
    Slurm, IncompleteVar, Val, OutputVar, \
    ValidationError, FileSet, PythonCall, Wait, SubPipeline

from dry_pipe.task import Task, TaskStep
from dry_pipe.task_state import TaskState


class DryPipe:

    annotated_python_task_by_name = {}

    @staticmethod
    def python_call(tests=[]):
        """
        annotation for methods that are called by Tasks, i.e. to go in a task's calls(...) clause
        :param tests: test cases
        A test case is a dict where keys map to arguments of the function.

        .. highlight:: python
        .. code-block:: python
            @DryPipe.python_call(test=[{"a":123, "b": "abc"}, {"a": 987, "b": "z"}])
            def my_func(a, b, test=None):
                if test is not None:
                    print(f"I'm being tested, a={a} and b={b}")
                else:
                    print("normal call")

        Ex, to run the 2nd test above with the CLI:

        .. highlight:: shell
        .. code-block:: shell
            drypipe test <module>.my_func:1

        THe above call is equivalent to:
        .. highlight:: python
        .. code-block:: python
            t = {"a": 987, "b": "z"}
            my_func(**t, t)


        where idx is
        """
        return lambda func: PythonCall(func, tests)

    @staticmethod
    def create_pipeline(
        generator_of_tasks,
        pipeline_code_dir=None,
        task_conf=None,
        containers_dir=None,
        env_vars=None,
        remote_task_confs=None,
        task_groupers={
            "by_task_type": Task.key_grouper
        },
        pipeline_code_dir_ls_command=None
    ):
        """
        :param generator_of_tasks: a python generator function that generates Tasks
        :param pipeline_code_dir: The code source root of the pipeline. Defaults to the directory containing the python file containing the DAG generator
        :param task_conf:
        :param containers_dir:
        :param env_vars:
        :param remote_task_confs: when a list of TaskConf is given, drypipe prepare-remote-sites will
               upload (rsync) $containers_dir and $pipeline_code_dir to all remote sites
        :param task_groupers:
        .. highlight:: python
        .. code-block:: python

            DryPipe.create_pipeline(
                my_dag_generator,
                task_groupers={
                    "group_by_task_key_last_char": lambda task_key: task_key[-1]
                }
            )
        """

        from dry_pipe.pipeline import Pipeline

        return Pipeline(
            generator_of_tasks, pipeline_code_dir, task_conf, containers_dir, env_vars, remote_task_confs,
            task_groupers, pipeline_code_dir_ls_command
        )

    @staticmethod
    def pipeline_code_dir_for(task_generator_func):
        return os.path.dirname(os.path.abspath(inspect.getmodule(task_generator_func).__file__))


class TaskSetAccessor:

    def __init__(self, task_match_all, product_var):
        self.task_match_all = task_match_all
        self.product_var = product_var

    def as_glob_expression(self):
        return Val(
            os.path.join(
                "$__pipeline_instance_dir",
                "output",
                self.task_match_all.task_match.pattern,
                self.product_var.file_path
            ),
            is_glob_expression=True
        )

    def fetch(self):
        return [
            task.out.__getattr__(self.product_var.name).fetch()
            for task in self.task_match_all.task_match.tasks
        ]


class TaskMatchAll:
    def __init__(self, task_match):
        self.task_match = task_match

    def __getattr__(self, name):
        task = self.task_match.tasks[0]
        product_var = task.out.__getattr__(name)
        if product_var is None:
            raise Exception(f"task {task}, does not produce '{product_var}'")

        return TaskSetAccessor(self, product_var)


class TaskMatch:
    def __init__(self, pattern, tasks):
        self.pattern = pattern
        self.tasks = tasks
        self.all = TaskMatchAll(self)


class DryPipeDsl:

    def __init__(self, task_by_keys={}, task_conf=None, pipeline_instance=None, task_namespance_prefix=""):

        self.pipeline_instance = pipeline_instance
        self.task_conf = task_conf or TaskConf("process")
        self.task_namespance_prefix = task_namespance_prefix
        self.task_by_keys = task_by_keys

    def file_in_pipeline_instance_dir(self, file_name, must_exist=True):
        f = os.path.join(self.pipeline_instance.pipeline_instance_dir, file_name)
        if must_exist and not os.path.exists(f):
            raise Exception(f"file not found {f}")
        return f

    def sub_pipeline(self, pipeline, namespace_prefix):
        """
        :param pipeline: the sub pipeline
        :param namespace_prefix: The prefix for all task keys of the sub pipeline
        :return: an instance of SubPipeline, must be yielded
        """
        return SubPipeline(pipeline, namespace_prefix, self)

    def wait_for_matching_tasks(self, *patterns):
        """

        :param patterns: one or more [glob pattern](https://docs.python.org/3/library/glob.html)
        :return: a tuple of one :py:class:<dry_pipe.TaskMatch> for every pattern passed as argument

        ex:
            for matcher_for_a, matcher_for_b in dsl.wait_for_matching_tasks("a*","b*"):
        """

        task_matchers = []
        current_pattern_idx = 0

        for pattern in patterns:

            tasks = []

            for key, task in self.task_by_keys.items():
                if fnmatch(key, pattern):
                    self.pipeline_instance.dag_determining_tasks_ids.add(key)
                    if not task.has_completed():
                        return []
                    else:
                        tasks.append(task)

            if len(tasks) == 0:
                if current_pattern_idx == 0:
                    raise Exception(f"with_completed_matching_tasks({pattern}) matched zero tasks")
                else:
                    return []

            task_matchers.append(TaskMatch(pattern, tasks))
            current_pattern_idx += 1

        if len(task_matchers) == 1:
            yield task_matchers[0]
        else:
            yield tuple(task_matchers)

    def wait_for_tasks(self, *args):
        """

        :param args: one or more task, or task_key
        :return: a tuple of one task for every task or task key passed as argument
        ex1: for t1 in dsl.wait_for_tasks(t1):

        ex2: for t1, t2, t3 in dsl.wait_for_tasks(task1, task2, "key_of_task3"):
                assert t1 == task1
                assert t2 == task2
                assert t3.key == "key_of_task3"

        Note: t1
        """

        def it_completed_tasks():
            for a in args:
                if isinstance(a, str):
                    key = a
                elif isinstance(a, Task):
                    key = a.key
                else:
                    raise ValidationError(
                        f"illegal argument {type(a)} given to with_tasks, must be Task or a string Task key"
                    )

                self.pipeline_instance.dag_determining_tasks_ids.add(key)

                t = self.task_by_keys.get(key)

                if t is None:
                    break
                else:
                    state = t.get_state()
                    if state is not None and state.is_completed():
                        yield t
                    else:
                        break

        number_of_tasks = len(args)

        if number_of_tasks == 0:
            raise ValidationError(f"must supply at least one task key")

        completed_tasks = list(it_completed_tasks())

        if number_of_tasks == len(completed_tasks):
            if number_of_tasks == 1:
                yield completed_tasks[0]
            else:
                yield tuple(completed_tasks)

    def var(self, type=str, may_be_none=False):
        """
        :param type: str, int, or float
        :param may_be_none:
        :return an object that goes in task.produces():
        """
        return IncompleteVar(type, may_be_none)

    def val(self, v):
        """
        constant value to pass a task input, in the consumes(...) clause
        :param v: the value, either an int or a str
        :return:
        """
        return Val(v)

    def file(self, name, manage_signature=None, remote_cache_bucket="pipeline-instance"):

        if remote_cache_bucket is not None:
            legal_values = ["pipeline-instance", "task"]
            if remote_cache_bucket not in legal_values:
                raise Exception(f"remote_cache_bucket can't be {remote_cache_bucket}, must be one of {legal_values}")

        if type(name) != str:
            raise ValidationError(f"invalid file name, must be a string {name}")

        return IndeterminateFile(name, manage_signature, remote_cache_bucket)

    def fileset(self, glob_pattern):

        return FileSet(glob_pattern)

    def task(self,
             key,
             task_conf=None):
        """
        Creates a :py:meth:`dry_pipe.TaskBuilder`, the object for declaring tasks

        .. highlight:: python
        .. code-block:: python

            def my_dag_generator(dsl):
                yield dsl.task(key="t1").\\
                    consumes(other_task.out.x).\\
                    produces(z=dsl.var(int)).\\
                    calls(f1)()
        :param key:
        :param task_conf:
        :return: TaskBuilder
        """

        if task_conf is None:
            task_conf = self.task_conf

        if key is None or key == "":
            raise Exception(f"invalid key given to task(...): '{key}'")

        key = f"{self.task_namespance_prefix}{key}"

        return TaskBuilder(
            key=key,
            _produces={},
            _consumes={},
            dsl=self,
            task_conf=task_conf,
            pipeline_instance=self.pipeline_instance
        )


class TaskBuilder:


    def __init__(self, key, _consumes={}, _produces={},
                 dsl=None, task_steps=[],
                 _upstream_task_completion_dependencies=None, _props=None, task_conf=None, pipeline_instance=None):

        self.key = key
        self.dsl = dsl
        self._props = _props or {}
        self._consumes = _consumes
        self._upstream_task_completion_dependencies = _upstream_task_completion_dependencies or []
        self._produces = _produces
        self.task_steps = task_steps
        self.task_conf = task_conf
        self.pipeline_instance = pipeline_instance

    def _deps_from_kwargs(self, kwargs):

        def deps():
            for k, v in kwargs.items():
                if isinstance(v, IndeterminateFile):
                    yield k, v
                elif isinstance(v, ProducedFile) or isinstance(v, Val) or isinstance(v, OutputVar):
                    yield k, v
                else:
                    raise ValidationError(
                        f"_consumes can only take DryPipe.file() or _consumes(a_file=other_task.out.name_of_file()" +
                        f"task(key={self.key}) was given {type(v)}",
                        ValidationError.consumes_has_invalid_kwarg_type
                    )
        return {
            ** self._consumes,
            ** dict(deps())
        }

    def consumes(self, *args, **kwargs):
        """
        The consumes clause
        :param args:
        :param kwargs:
        :return a new :py:meth:`dry_pipe.TaskBuilder` with the added consumes declaration
        """

        def deps_from_args():
            for o in args:
                if isinstance(o, ProducedFile):
                    yield o.var_name, o
                elif isinstance(o, OutputVar):
                    yield o.name, o
                else:
                    raise ValidationError(
                        f"bad arg: {o} passed to {self}"
                    )

        for v, count in collections.Counter([k for k, _ in deps_from_args()]).items():
            if count > 1:
                raise Exception(f"duplicate variable {v} in task(key={self.key}).consumes(...) clause")

        return TaskBuilder(** {
            ** vars(self),
            ** {"_consumes": {
                    ** self._consumes,
                    ** dict(deps_from_args()),
                    ** dict(self._deps_from_kwargs(kwargs))
                }
            }
        })

    def props(self, **kwargs):
        return TaskBuilder(** {
            ** vars(self),
            ** {"_props": kwargs}
        })

    def produces(self, *args, **kwargs):
        """
        The produces clause
        :param args:
        :param kwargs:
        :return a new :py:meth:`dry_pipe.TaskBuilder` with the added produces declaration
        """

        if len(args) > 0:
            raise ValidationError(
                f"DryPipe.produces(...) can't take positional args, use the form produce(var_name=...)",
                ValidationError.produces_cant_take_positional_args
            )

        def outputs():
            for k, v in kwargs.items():
                if isinstance(v, IndeterminateFile):
                    yield k, v
                elif isinstance(v, IncompleteVar):
                    yield k, v
                elif isinstance(v, FileSet):
                    yield k, v
                else:
                    raise ValidationError(
                        f"produces takes only DryPipe.file or DryPipe.vars, ex:\n " +
                        "1:    task(...).produces(var_name=DryPipe.file('abc.tsv'))\n"
                        " 2:    task(...).produces(vars=DryPipe.vars(x=123,s='a'))",
                        ValidationError.produces_only_takes_files
                    )

        return TaskBuilder(** {
            ** vars(self),
            ** {"_produces": dict(list(outputs()))}
        })

    def calls(self, *args, **kwargs):
        """
        Adds a calls() to the task declaration
        :param args: a python function annotated with :py:meth:`.python_call` or a bash snippet with a proper shebang
        :return a new :py:meth:`dry_pipe.TaskBuilder` with the added call
        """

        task_conf = self.task_conf
        container = kwargs.get("container")
        if container is not None:
            task_conf = task_conf.override_container(container)

        executer_type = kwargs.get("executer_type")
        if executer_type is not None:
            task_conf = task_conf.override_executer(executer_type)

        task_step = None

        if len(args) == 1:
            a = args[0]
            if type(a) == str:
                if a.endswith(".sh"):
                    task_step = TaskStep(task_conf, shell_script=a)
                else:
                    script_text = textwrap.dedent(a)
                    if re.match("\\n(\\w*)#!/.*", script_text):
                        start_idx = script_text.find("#!")
                        task_step = TaskStep(task_conf, shell_snippet=script_text[start_idx:-1])
                    else:
                        raise ValidationError(
                            f"invalid arg to clause:\n ...calls({a})\nvalid arg is a script file (.sh suffix), " +
                            "or a code block with shebang (ex):\n" +
                            f"{bash_shebang}"
                            "echo '...something...'"
                        )
            if isinstance(a, PythonCall):
                python_bin = kwargs.get("python_bin") or self.dsl.task_conf.python_bin or sys.executable
                task_conf = task_conf.override_python_bin(python_bin)
                task_step = TaskStep(task_conf, python_call=a)

        if task_step is None:
            raise ValidationError(
                f"invalid args, task.calls(...) can take a sigle a single positional argument, was given: {args}",
                ValidationError.call_has_bad_arg
            )

        return TaskBuilder(** {
            ** vars(self),
            "task_steps": self.task_steps + [task_step]
        })

    def __call__(self):
        return Task(self)


def host_has_sbatch():

    with PortablePopen(
        ["which", "sbatch"]
    ) as p:

        p.wait()

        if p.returncode != 0:
            return False

        out = p.stdout.read().strip()

        if out == "":
            return False

        if out.endswith("/sbatch"):
            return True

        raise Exception(f"Funny return from which sbatch: {out}")


class TaskConf:
    """
    :param executer_type: 'process' or 'slurm'
    :param ssh_specs: me@a-host:PORT:/path/to/ssh-private-key
    :param slurm_account: a slurm username
    :param sbatch_options: list of string for every sbatch option, ex: ["--time=0:1:00", "--mem=30G", "--cpus-per-task=24"]
    :param container: the file name of the container, ex: my-container.sif (without path, containers
        the file must exist in
          $__containers_dir/<container>
        the default value of $__containers_dir is $__pipeline_code_dir/containers
        Note: for remote tasks (when ssh_specs is defined, the container file must exist in `remote_containers_dir`)
    :param command_before_launch_container:
    :param remote_pipeline_code_dir:
    :param python_bin:
     By default, will use the same python bin as the one running the cli
     to run in a specific virtualenv:
        /path/to/my_virtualenv/bin/python
     to run the task in a miniconda env:
        /path/to/my-miniconda3/envs/<my-conda-env>/bin/python
     if running in a container, it should point to the python executable in the container, most likel:
        /usr/bin/python3
    :param remote_base_dir:
        the remote directory that will contain the $__pipeline_instance_dir structure
    :param remote_containers_dir:
        the directory containing the container sif files
    :param python_interpreter_switches:
        extra switches to add to the python executable that will launch the python call (only applies to PythonCall)
    """

    @staticmethod
    def default():
        return TaskConf("process")

    def __init__(
            self,
            executer_type=None,
            ssh_specs=None,
            slurm_account=None,
            sbatch_options=[],
            container=None,
            command_before_launch_container=None,
            remote_pipeline_code_dir=None,
            python_bin=None,
            remote_base_dir=None,
            remote_containers_dir=None,
            init_bash_command=None,
            python_interpreter_switches=["-u"],
            fields_from_json=None,
            extra_env=None,
            label=None
    ):

        if executer_type is None:
            executer_type = "process"

        if executer_type not in ["slurm", "process"]:
            raise Exception(f"invalid executer_type: {executer_type}")

        if executer_type == "slurm":
            if slurm_account is None:
                raise Exception("slurm_account must be specified when executer_type is slurm")
            if slurm_account == "":
                raise Exception("slurm_account can't be ''")

        if executer_type == "process" and slurm_account is not None:
            raise Exception(f"can't specify slurm_account when executer_type is not 'slurm'")

        if fields_from_json is not None:
            self.__dict__.update(fields_from_json)
            return

        self.executer_type = executer_type
        self.ssh_specs = ssh_specs
        self.slurm_account = slurm_account
        self.sbatch_options = sbatch_options
        self.container = container
        self.command_before_launch_container = command_before_launch_container
        self.remote_pipeline_code_dir = remote_pipeline_code_dir
        self.python_bin = python_bin
        self.remote_base_dir = remote_base_dir
        self.remote_containers_dir = remote_containers_dir
        self.init_bash_command = init_bash_command
        self.python_interpreter_switches = python_interpreter_switches
        self.extra_env = extra_env
        self.label = label

        if extra_env is not None:
            if not isinstance(extra_env, dict):
                raise Exception(f"extra_env must be a dict of strings, got {type(extra_env)}")
            for k, v in extra_env.items():
                if not isinstance(v, str):
                    raise Exception(f"invalid value given to extra_env['{k}']: {type(v)}, must be str")

        if self.python_bin is None:
            if self.is_remote():
                self.python_bin = "/usr/bin/python3"
            else:
                self.python_bin = sys.executable

        if self.ssh_specs is not None:
            self.ssh_username, self.ssh_host, self.key_filename = parse_ssh_specs(self.ssh_specs)

            self.remote_site_key = ":".join([
                self.ssh_username,
                self.ssh_host,
                self.remote_base_dir
            ])

    def full_remote_path(self, pipeline_instance):
        return f"{self.remote_site_key}/{pipeline_instance.instance_dir_base_name()}"

    def as_json(self):
        return dict(
            (key, value)
            for key, value in self.__dict__.items() if not callable(value) and not key.startswith('__')
        )

    @staticmethod
    def from_json(json_dict):
        return TaskConf(fields_from_json=json_dict)

    @staticmethod
    def from_json_file(json_file):
        with open(json_file) as f:
            return TaskConf.from_json(json.loads(f.read()))

    def is_remote(self):
        return self.ssh_specs is not None

    def is_slurm(self):
        return self.executer_type == "slurm"

    def is_process(self):
        return self.executer_type == "process"

    def has_container(self):
        return self.container is not None

    def _ensure_is_remote(self):
        if not self.is_remote():
            raise Exception(f"can't call this on this non remote TaskConf")

    def uses_singularity(self):
        return self.container is not None

    def override_container(self, container):
        return TaskConf(
            self.executer_type,
            self.ssh_specs,
            self.slurm_account,
            self.sbatch_options,
            container,
            self.command_before_launch_container,
            self.remote_pipeline_code_dir,
            self.python_bin,
            self.remote_base_dir,
            self.remote_containers_dir,
            self.init_bash_command,
            self.python_interpreter_switches,
            extra_env=self.extra_env
        )

    def override_python_bin(self, python_bin):
        return TaskConf(
            self.executer_type,
            self.ssh_specs,
            self.slurm_account,
            self.sbatch_options,
            self.container,
            self.command_before_launch_container,
            self.remote_pipeline_code_dir,
            python_bin,
            self.remote_base_dir,
            self.remote_containers_dir,
            self.init_bash_command,
            self.python_interpreter_switches,
            extra_env=self.extra_env
        )

    def override_executer(self, executer_type):
        return TaskConf(
            executer_type,
            self.ssh_specs,
            self.slurm_account,
            self.sbatch_options,
            self.container,
            self.command_before_launch_container,
            self.remote_pipeline_code_dir,
            self.python_bin,
            self.remote_base_dir,
            self.remote_containers_dir,
            self.init_bash_command,
            self.python_interpreter_switches,
            extra_env=self.extra_env
        )

    _remote_ssh_executers = {}

    @staticmethod
    def _get_or_create_remote_ssh(ssh_username, ssh_host, remote_base_dir, key_filename, command_before_launch_container):
        from dry_pipe.ssh_executer import RemoteSSH

        return RemoteSSH(ssh_username, ssh_host, remote_base_dir, key_filename, command_before_launch_container)

    def create_executer(self):

        def remote_ssh():
            from dry_pipe.ssh_executer import RemoteSSH

            if self.remote_base_dir is None:
                raise Exception("A task_conf with ssh must have remote_base_dir not None")

            return RemoteSSH(
                self.ssh_username, self.ssh_host, self.remote_base_dir, self.key_filename,
                self.command_before_launch_container
            )

        if self.executer_type == "process":
            if self.is_remote():
                return remote_ssh()
            else:
                return Local(self.command_before_launch_container)
        else:
            if self.is_remote():
                e = remote_ssh()
                e.slurm = Slurm(self.slurm_account, self.sbatch_options)
                return e
            else:
                return Slurm(self.slurm_account, self.sbatch_options)



import os
import signal
import textwrap
import time
import unittest
from pathlib import Path
from threading import Thread

from dry_pipe import bash_shebang, script_lib
from dry_pipe.script_lib import task_script_header, touch, PortablePopen
from test_utils import TestSandboxDir


def _run_script(script, send_signal=None, env=None):

    with PortablePopen([script, "--is-silent", "--wait"], env=env) as p:

        if send_signal is not None:
            send_signal(p)

        p.popen.wait()
        out = p.stdout_as_string()
        err = p.stderr_as_string()
        return p.popen.returncode, out, err



class ScriptLibTests(unittest.TestCase):

    def _out_err_content(self, control_dir):
        with open(os.path.join(control_dir, "out.log")) as out:
            with open(os.path.join(control_dir, "err.log")) as err:
                return out.read(), err.read()

    def _write_task_env_file(self, control_dir, pseudo_task_key):
        task_env_file = os.path.join(control_dir, 'task-env.sh')
        with open(task_env_file, 'w') as f:
            f.write(bash_shebang())
            f.write(textwrap.dedent(f"""
                __script_location=$( cd "$( dirname "${'{BASH_SOURCE[0]}'}" )" >/dev/null 2>&1 && pwd )
                export __pipeline_instance_dir=$( dirname $( dirname $__script_location))
                export export __pipeline_code_dir=$__pipeline_instance_dir                
                # . $__pipeline_instance_dir/.drypipe/pipeline-env.sh
                export __control_dir=$__pipeline_instance_dir/.drypipe/{pseudo_task_key}
                export __out_log=$__control_dir/out.log
                export __err_log=$__control_dir/err.log
                export __task_output_dir=$__pipeline_instance_dir/output/{pseudo_task_key}   
                export __file_list_to_sign=$__task_output_dir/a.txt,$__task_output_dir/b.txt
            """))
        os.chmod(task_env_file, 0o764)

    def _single_step_bash_template(self, d, script_code):
        d.init_pid_for_tests()

        script_path = os.path.join(d.sandbox_dir, 'script.sh')

        with open(script_path, 'w') as f:
            f.write(script_code)

        pseudo_task_key = "task0"
        control_dir = os.path.join(d.sandbox_dir, '.drypipe', pseudo_task_key)
        Path(control_dir).mkdir()
        work_dir = os.path.join(d.sandbox_dir, 'output', pseudo_task_key)
        Path(work_dir).mkdir()
        self._write_task_env_file(control_dir, pseudo_task_key)
        os.chmod(script_path, 0o764)

        task_script = os.path.join(control_dir, 'task')

        with open(task_script, 'w') as f:
            f.write(task_script_header())
            f.write(textwrap.dedent(
                f"""
                env = script_lib.source_task_env(os.path.join(__script_location, 'task-env.sh'))                                        
                script_lib.touch(os.path.join(env['__control_dir'], 'output_vars'))                                                
                step_number, control_dir, state_file, read_task_state = script_lib.read_task_state()                                
                        
                state_file, step_number = script_lib.transition_to_step_started(state_file, step_number)                                        
                script_lib.run_script(os.path.join(env['__pipeline_code_dir'], 'script.sh'))                    
                state_file, step_number = script_lib.transition_to_step_completed(state_file, step_number)
                script_lib.sign_files()                                                        
                """
            ))
        os.chmod(task_script, 0o764)

        state_file = os.path.join(control_dir, 'state.queued.0')
        touch(state_file)

        return control_dir, task_script

    def test_bash_fail(self):
        d = TestSandboxDir(self)
        script_code = textwrap.dedent("""
            #!/usr/bin/env bash            
            echo "the end"
            echo "err99999" >&2
            exit 1            
        """).strip()

        control_dir, task_script = self._single_step_bash_template(d, script_code)

        return_code = _run_script(task_script)

        self.assertNotEqual(return_code, 0)

        out, err = self._out_err_content(control_dir)

        self.assertEqual(out, "the end\n")
        self.assertEqual(err, "err99999\n")

        step_number, control_dir, state_file, state_name = script_lib.read_task_state(control_dir)

        self.assertEqual(step_number, 0)
        self.assertEqual(os.path.basename(state_file), "state.failed.0")

    def _test_bash_timeout(self):
        d = TestSandboxDir(self)
        script_code = textwrap.dedent("""
            #!/usr/bin/env bash            
            echo "the end"
            echo "err99999" >&2
            for i in $(seq 1000000); do
                echo "--:>$i"
                sleep 2
            done                        
        """).strip()

        control_dir, task_script = self._single_step_bash_template(d, script_code)

        def send_signal(p):
            def f():
                time.sleep(1)
                p.send_signal(signal.SIGUSR1)
                time.sleep(1)
                p.kill()
            t = Thread(target=f)
            t.start()

        return_code = _run_script(task_script, send_signal)

        self.assertNotEqual(return_code, 0)

        out, err = self._out_err_content(control_dir)

        self.assertEqual(err, "err99999\n")

        step_number, control_dir, state_file, state_name = script_lib.read_task_state(control_dir)

        self.assertEqual(step_number, 0)
        self.assertEqual(os.path.basename(state_file), "state.timed-out.0")



import unittest

from pipelines_with_sub_pipelines import sum_and_multiply_by_x_and_add_y, sum_and_multiply_by_x_and_add_y_and_z
from test_utils import TestSandboxDir


class PipelineCompositionTests(unittest.TestCase):

    def test_composed_pipeline_2_levels(self):

        pipeline_instance = TestSandboxDir(self).pipeline_instance_from_generator(
            sum_and_multiply_by_x_and_add_y([1, 2, 3, 4], 5, 6),
            completed=True
        )

        self.assertEqual(
            int(pipeline_instance.tasks["add_y"].out.final_result.fetch()),
            sum([1, 2, 3, 4]) * 5 + 6
        )

    def test_composed_pipeline_3_levels(self):

        pipeline_instance = TestSandboxDir(self).pipeline_instance_from_generator(
            sum_and_multiply_by_x_and_add_y_and_z([1, 2, 3, 4], 5, 6, 17),
            completed=True
        )

        self.assertEqual(
            int(pipeline_instance.tasks["add_z"].out.final_result.fetch()),
            sum([1, 2, 3, 4]) * 5 + 6 + 17
        )



import unittest

import os
import test_helpers
from dry_pipe import TaskConf

from tests.pipeline_with_multistep_tasks import three_steps_pipeline_task_generator, \
    hybrid_bash_python_mutlistep_pipeline_task_generator, three_steps_pipeline_expected_output
from test_utils import TestSandboxDir


class MultipstepTaskTests(unittest.TestCase):

    def test_normal_execution(self):

        d = TestSandboxDir(self)

        pipeline_instance = d.pipeline_instance_from_generator(
            three_steps_pipeline_task_generator,
            completed=True
        )

        three_phase_task = next(pipeline_instance.tasks_for_key_prefix("three_phase_task"))

        three_step_task_out_file = os.path.join(three_phase_task.v_abs_work_dir(), "out_file.txt")

        def load_output_from_three_phase_task():
            return three_phase_task.out.out_file.load_as_string()

        self.assertEqual(three_steps_pipeline_expected_output(), three_phase_task.out.out_file.load_as_string())

        pipeline_instance.clean_all()

        from dry_pipe.internals import Local
        try:
            Local.fail_silently_for_test = True

            pipeline_instance.run_sync(tmp_env_vars={
                "CRASH_STEP_2": "true"
            })

            self.assertEqual("s1\n", load_output_from_three_phase_task())
            self.assertTrue(three_phase_task.get_state().is_failed())

            test_helpers.rewrite_file(three_step_task_out_file, "")

            three_phase_task.re_queue()

            self.assertFalse(three_phase_task.get_state().is_failed())

            pipeline_instance.run_sync(tmp_env_vars={
                "CRASH_STEP_3": "true"
            })

            # ensure s1 is not there, i.e. step1 has NOT executed
            self.assertEqual("s2\n", load_output_from_three_phase_task())

            three_phase_task.re_queue()

            pipeline_instance.run_sync()

            self.assertEqual("s2\ns3\n", load_output_from_three_phase_task())

            """
            event_lines = list(three_phase_task.parse_tracking_file())
    
            step_lines = [
                int(l.split("_")[1])
                for l in event_lines if l.startswith("step_")
            ]
    
            self.assertEqual(step_lines, [0, 0, 1, 1, 2, 2])
    
            event_lines = [
                l.split("=")[0]
                for l in event_lines
            ]
    
            self.assertTrue("launched_at" in event_lines)
            self.assertTrue("completed_at" in event_lines)
            """
        finally:
            Local.fail_silently_for_test = False

    def test_pipeline_with_mixed_python_bash(self):

        pipeline = TestSandboxDir(self).pipeline_instance_from_generator(
            hybrid_bash_python_mutlistep_pipeline_task_generator,
            completed=True,
            task_conf=TaskConf(
                executer_type="process"
                #python_bin="/usr/bin/python3"
            )
        )

        three_phase_task = next(pipeline.tasks_for_key_prefix("three_phase_task"))

        three_phase_task_out_file = os.path.join(three_phase_task.v_abs_work_dir(), "out_file.txt")

        self.assertEqual("s1\ns2\ns3\ns4\n", test_helpers.load_file_as_string(three_phase_task_out_file))



import glob
import inspect
import os
import subprocess
import getpass
import logging.config

from dry_pipe.script_lib import PortablePopen

LOCAL_PROCESS_IDENTIFIER_VAR = "____DRY_PIPE_TASK"
logger = logging.getLogger(__name__)


class ValidationError(Exception):
    def __init__(self, message, code=None):
        super(ValidationError, self).__init__(message)
        self.code = code

    consumes_has_bad_positional_arg_val = 0

    consumes_has_bad_positional_arg = 1

    consumes_has_invalid_kwarg_type = 2

    produces_cant_take_positional_args = 3

    produces_only_takes_files = 4

    call_has_bad_arg = 5


class Val:

    def __init__(self, value, is_glob_expression=False):
        self.is_glob_expression = is_glob_expression
        t = type(value)
        if t == int or t == str or t == float:
            self.value = value
            self.typez = t
        elif t == type(int):
            raise ValidationError(f"invalid val type dsl.val(int), you probably want to use 'dsl.var(int)'")
        elif t == t == type(str):
            raise ValidationError(f"invalid val type dsl.var(str), you probably want to use 'dsl.var(str)'")
        elif type(float):
            raise ValidationError(f"invalid val type, dsl.var(float) you probably want to use 'dsl.var(float)'")
        else:
            raise ValidationError(f"invalid val type: {type(value)}, must be int, string, float")

    def type_str(self):
        return _type_str(self.typez)

    def to_hash(self):
        return str(self.value)

    def serialized_value(self):

        if type(self.value) == str:
            return f'"{self.value}"'

        return f"{self.value}"


class IncompleteVar:

    def __init__(self, type=str, may_be_none=True):
        self.type = type
        self.may_be_none = may_be_none


def _type_str(t):
    if t == int:
        return "int"
    elif t == str:
        return "str"
    elif t == float:
        return "float"


class OutputVar:

    def __init__(self, type, may_be_none, name, producing_task):

        supported = [str, int, float]

        if type not in supported:
            raise ValidationError(
                f"var {name} in {producing_task} has invalid type. Supported types are: " +
                ', '.join([_type_str(t) for t in supported])
            )

        self.name = name
        self.type = type
        self.producing_task = producing_task
        self.may_be_none = may_be_none

    def type_str(self):
        return _type_str(self.type)

    def format_for_python(self, v):

        if self.type == int:
            return str(v)
        elif self.type == str:
            return f'"{v}"'
        elif self.type == float:
            return str(v)

        raise Exception(f"unsupported type {self.type}")

    def unformat_for_python(self, v):

        if self.type == int:
            return int(v)
        elif self.type == str:
            s = v.rstrip('\"\'')
            s = s.lstrip('\"\'')
            return s
        elif self.type == float:
            return float(v)

        raise Exception(f"unsupported type {self.type}")

    def input_var(self, var_name_in_consuming_task):
        return InputVar(self, var_name_in_consuming_task)

    def fetch(self):
        varz = self.producing_task.out_vars_by_name()
        v = varz.get(self.name)
        if v is None:
            return None
        return self.unformat_for_python(v)


class InputVar:

    def __init__(self, output_var, var_name_in_consuming_task):
        self.output_var = output_var
        self.var_name_in_consuming_task = var_name_in_consuming_task

    def write(self, string_value):
        return string_value


class Executor:

    def key(self):
        return None

    def is_remote(self):
        return False


def _process_iter(names):
    import psutil

    for it in psutil.process_iter(['pid', 'exe', 'username', 'environ']):
        yield it


def _is_access_denied(ex):
    from psutil import AccessDenied
    return isinstance(ex, AccessDenied)

class Local(Executor):

    def count_running_tasks(self):

        current_user = getpass.getuser()

        c = 0

        for it in _process_iter(['pid', 'exe', 'username', 'environ']):

            if it.username() != current_user:
                continue
            try:

                if not it.exe().endswith("/bash"):
                    continue

                v = it.environ().get(LOCAL_PROCESS_IDENTIFIER_VAR)

                if v is not None: # and v.startswith("____"):
                    c += 1

            except Exception as ex:
                if not _is_access_denied(ex):
                    raise ex

        return c

    def kill_task(self, task_key):

        current_user = getpass.getuser()

        for it in _process_iter(['pid', 'exe', 'username', 'environ']):

            if it.username() != current_user:
                continue
            try:

                if not it.exe().endswith("/bash"):
                    continue

                v = it.environ().get(LOCAL_PROCESS_IDENTIFIER_VAR)

                if v is not None:
                    if v == f"____{task_key}":
                        it.kill()

            except Exception as ex:
                if not _is_access_denied(ex):
                    raise ex

        #pids = list(get_pids())
        #print(f"pids: {pids}")
        #for pid in pids:
        #    os.kill(pid. signal.SIGKILL)

    def __init__(self, before_execute_bash):
        self.before_execute_bash = before_execute_bash

    fail_silently_for_test = False

    """
        experimental, for scenario:
        + launcher runs in a LXC container
        + launcher runs tasks with singularity,launch on the physical host of the LXC container  
    """
    def __command_on_sibling_container(self, task, user, back_ground):
        return [
            f"ssh {user}@127.0.0.1 'nohup {task.v_abs_script_file()} " +
            f"  >{task.v_abs_out_log()} 2>{task.v_abs_err_log()} {back_ground}'"
        ]

    def execute(self, task, touch_pid_file_func, wait_for_completion=False, fail_silently=False):
        if wait_for_completion:
            cmd = [task.v_abs_script_file(), "--wait"]
        else:
            cmd = [task.v_abs_script_file()]

        logger.debug("will launch task %s", ' '.join(cmd))

        with PortablePopen(
            cmd,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL
        ) as p:
            p.wait()
            if p.popen.returncode != 0:
                logger.warning("task ended with non zero code: %s", p.popen.returncode)
            else:
                logger.debug("task ended with returncode: %s", p.popen.returncode)


class Slurm(Executor):

    def __init__(self, account, sbatch_options):

        if account is None:
            raise Exception(f"account account can't be None")

        self.account = account
        self.sbatch_options = sbatch_options

    def execute(self, task, touch_pid_file_func, wait_for_completion=False, fail_silently=False):

        env = {
            **os.environ
        }

        if wait_for_completion:
            cmd = [f"bash -c 'export SBATCH_WAIT=True && {task.v_abs_sbatch_launch_script()}'"]
        else:
            cmd = [task.v_abs_sbatch_launch_script()]

        with PortablePopen(
            cmd,
            shell=True,
            env=env
        ) as p:
            p.wait_and_raise_if_non_zero()


class IndeterminateFile:

    def __init__(self, file_path, manage_signature, remote_cache_bucket=None):
        self.file_path = file_path.strip()
        self.manage_signature = manage_signature
        self.remote_cache_bucket = remote_cache_bucket

    def produced_file(self, var_name, producing_task):
        return ProducedFile(self.file_path, var_name, self.manage_signature, producing_task)

    def pre_existing_file(self):
        return PreExistingFile(self.file_path, self.manage_signature, self.remote_cache_bucket)


class FileSet:

    def __init__(self, glob_pattern):
        self.glob_pattern = glob_pattern

    def out_file_set(self, producing_task, name_in_producing_task):
        return OutFileSet(self, producing_task, name_in_producing_task)


class OutFileSet:

    def __init__(self, file_set, producing_task, name_in_producing_task):
        self.file_set = file_set
        self.producing_task = producing_task
        self.name_in_producing_task = name_in_producing_task

    def fetch(self):
        if not self.producing_task.get_state().is_completed():
            raise Exception(f"task must be complete before out value can be fetched")

        producing_task_work_dir = self.producing_task.v_abs_work_dir()

        pattern = os.path.join(producing_task_work_dir, self.file_set.glob_pattern)

        def make_path_relative_to_pipeline_instance_dir(p):
            return os.path.relpath(p, producing_task_work_dir)

        for f in glob.glob(pattern):
            f = make_path_relative_to_pipeline_instance_dir(f)
            yield ProducedFile(f, self.name_in_producing_task, True, self.producing_task, self.file_set.glob_pattern)


class SubPipeline:
    def __init__(self, pipeline, task_namespance_prefix, dsl):
        self.pipeline = pipeline
        self.task_namespance_prefix = task_namespance_prefix
        self.dsl = dsl

    def wait_for_tasks(self, *args):
        args = [f"{self.task_namespance_prefix}{a}" for a in args]
        return self.dsl.wait_for_tasks(*args)


class PreExistingFile:

    def __init__(self, file_path, manage_signature=False, remote_cache_bucket=None):

        if file_path is None or file_path == "" or file_path == ".":
            raise ValidationError(f"invalid file {file_path}")

        self.file_path = file_path
        self.manage_signature = manage_signature
        self.remote_cache_bucket = remote_cache_bucket

    def to_hash(self):
        return self.file_path

    def value(self):
        return self.file_path

    def absolute_path(self, task):

        if os.path.isabs(self.file_path):
            return self.file_path

        return self.file_path


class ProducedFile:

    def __init__(self, file_path, var_name, manage_signature, producing_task, is_dummy=False, glob_pattern=None):

        if file_path is None or file_path == "" or file_path == ".":
            raise ValidationError(f"invalid file {file_path}")

        if var_name is None or var_name == "":
            raise ValidationError(f"can't be {var_name}")

        self.var_name = var_name
        self.file_path = file_path
        self.producing_task = producing_task
        self.manage_signature = manage_signature
        self.is_dummy = is_dummy
        self.glob_pattern = glob_pattern

    def basename(self):
        return self.file_path

    def to_hash(self):
        return self.file_path

    def value(self):
        return self.file_path

    def absolute_path(self, task):

        if os.path.isabs(self.file_path):
            return self.file_path

        return os.path.join(task.work_dir(), self.file_path)

    def load_as_string(self):
        f = os.path.join(
            self.producing_task.pipeline_instance.pipeline_instance_dir,
            self.absolute_path(self.producing_task)
        )
        with open(f) as f:
            return f.read()

    def input_file(self, var_name_in_consuming_task):
        return InputFile(self, var_name_in_consuming_task)


class InputFile:

    def __init__(self, produced_file, var_name_in_consuming_task):
        self.produced_file = produced_file
        self.var_name_in_consuming_task = var_name_in_consuming_task


class PythonCall:

    def __init__(self, func, tests=[]):
        self.func = func
        self.signature = inspect.signature(self.func)
        self.tests = tests

    def signature_spec(self):
        raise Exception(f"implement me")

    def mod_func(self):
        mod = inspect.getmodule(self.func)
        file_name = os.path.basename(mod.__file__)
        func_name = self.func.__name__
        importable_module_name = f"{mod.__package__}.{file_name[:-3]}:{func_name}"
        return importable_module_name


class Wait:

    def __init__(self, tasks):
        self.tasks = tasks

    def is_ready(self):
        for t in self.tasks:
            if not t.has_completed():
                return False

        return True

class TaskProps:

    def __init__(self, task, props):
        self.props = props
        self.task = task

    def __getattr__(self, name):
        if name not in self.props:
            raise ValidationError(f"{self.task} has no prop {name}")
        return self.props.get(name)


def flatten(the_list):
    return [item for sublist in the_list for item in sublist]


class MessageUtils:

    @staticmethod
    def safe_utf8(s):
        try:
            return s.decode("utf8")
        except UnicodeDecodeError as ex:
            return str(s)




import glob
import importlib
import inspect
import json
import logging
import logging.config
import pathlib
import re
import signal
import sys
import os
import time
import traceback

import click
from dry_pipe import DryPipe
from dry_pipe.internals import PythonCall
from dry_pipe.janitors import Janitor
from dry_pipe.monitoring import fetch_task_groups_stats
from dry_pipe.pipeline import PipelineInstance, Pipeline
from dry_pipe.pipeline_state import PipelineState
from dry_pipe.script_lib import env_from_sourcing, parse_in_out_meta, create_task_logger, iterate_out_vars_from, \
    write_out_vars
from dry_pipe.task_state import NON_TERMINAL_STATES

logger = logging.getLogger(__name__)


@click.group()
@click.option('-v', '--verbose')
@click.version_option()
@click.pass_context
def cli_group(ctx, verbose):
    ctx.ensure_object(dict)
    ctx.obj['verbose'] = verbose
    ctx.obj["ctx"] = ctx


def _bad_entrypoint(ctx):

    if "entrypoint" in ctx.obj and ctx.obj["entrypoint"] == "run_task":
        click.echo("Can't launch a pipeline with cli.run_task(), use cli.run(pipeline=<your pipeline>)", err=True)
        return True

    return False


def _configure_logging(log_conf_file):

    if log_conf_file is None:
        log_conf_file = os.environ.get("DRYPIPE_LOGGING_CONF")

    loggin_conf_in_current_dir = os.path.join(os.getcwd(), "drypipe-logging-conf.json")

    if log_conf_file is None and os.path.exists(loggin_conf_in_current_dir):
        log_conf_file = loggin_conf_in_current_dir

    if log_conf_file is not None:
        with open(log_conf_file) as f:
            conf = json.loads(f.read())

            handlers = conf["handlers"]

            for k, handler in handlers.items():
                handler_class = handler["class"]
                if handler_class == "logging.FileHandler":
                    filename = handler.get("filename")
                    if filename is None:
                        raise Exception(f"logging.FileHandler '{k}' has no filename attribute in {log_conf_file}")
                    if "$" in handler["filename"]:
                        raise Exception(
                            f"filename attribute of '{k}': {filename} has undefined env variables, in {log_conf_file}"
                        )
                    handler["filename"] = os.path.expandvars(filename)

            logging.config.dictConfig(conf)
        logger.info("using logging conf: %s", log_conf_file)


@click.command()
@click.pass_context
@click.option('--instance-dir', type=click.Path(), default=None)
def clean(ctx, instance_dir):
    if _bad_entrypoint(ctx):
        return
    pipeline_func = ctx.obj["pipeline_func"]
    pipeline = _pipeline_from_pipeline_func(pipeline_func, instance_dir)
    # TODO: prompt
    pipeline.clean()


def _pipeline_from_pipeline_func(pipeline_func, instance_dir):

    pipeline_func_sig = inspect.signature(pipeline_func)

    nargs = len(pipeline_func_sig.parameters)

    if nargs == 0:

        if instance_dir is not None:
            raise Exception(f"when specifying instance_dir, task generator must take a dry_pipe.DryPipe.dsl() argument")

        return DryPipe.create_pipeline(pipeline_func)
    elif nargs == 1:
        #dsl = DryPipe.dsl_for(pipeline_func)
        pipeline_code_dir = DryPipe.pipeline_code_dir_for(pipeline_func)
        pipeline_instance_dir = pipeline_code_dir

        #if instance_dir is not None:
        #    dsl = dsl.with_defaults(pipeline_instance_dir=pipeline_code_dir)
        return DryPipe.create_pipeline(pipeline_func, pipeline_code_dir)
    else:
        m = inspect.getmodule(pipeline_func)
        raise Exception(f"function {pipeline_func.__name__} in {m.__file__} takes {nargs} arguments,\n"
                        f"it should take a instance created with dry_pipe.DryPipe.dsl()")


@click.command()
@click.pass_context
@click.option('--clean', is_flag=True)
@click.option('--instance-dir', type=click.Path(), default=None)
def prepare(ctx, clean, instance_dir):

    if _bad_entrypoint(ctx):
        return

    pipeline_func = ctx.obj["pipeline_func"]

    pipeline = _pipeline_from_pipeline_func(pipeline_func, instance_dir)

    if clean:
        pipeline.clean_all()

    pipeline.init_work_dir()

    pipeline.get_state().touch()

    tasks_total = 0
    work_done = 0
    tasks_in_non_terminal_states = 0
    tasks_completed = 0

    for task in pipeline.tasks:

        tasks_total += 1

        task_state = task.get_state()

        if task_state is None:
            task.create_state_file_and_control_dir()
            work_done += 1
            task_state = task.get_state()

        if task_state.state_name in NON_TERMINAL_STATES:
            tasks_in_non_terminal_states += 1

        if task_state.is_waiting_for_deps():

            if not task.has_unsatisfied_deps():
                if task_state.is_prepared():
                    continue
                else:
                    task_state.transition_to_prepared(task)
                    task_state = task.get_state()
                    task_state.transition_to_queued(task)
                    task.prepare()
                    tasks_in_non_terminal_states += 1
                    work_done += 1

        elif task_state.is_completed():
            if task_state.action_if_exists() is None:
                tasks_completed += 1

    if tasks_total == tasks_completed:
        pipeline_state = pipeline.get_state()
        if not pipeline_state.is_completed():
            pipeline_state.transition_to_completed()


def _pipeline_instance_creater(instance_dir, module_func_pipeline, env_vars):

    if module_func_pipeline is None:
        module_func_pipeline = PipelineInstance.guess_pipeline_from_hints(instance_dir)

    pipeline = Pipeline.load_from_module_func(module_func_pipeline)

    PipelineInstance.write_hint_file_if_not_exists(instance_dir, module_func_pipeline)

    return lambda: pipeline.create_pipeline_instance(instance_dir, env_vars=env_vars)


@click.command()
@click.option('-p', '--pipeline', help="a_module:a_func, a function that returns a pipeline.")
@click.option('--instance-dir', type=click.Path(), default=None)
@click.option(
    '--group-by', type=click.STRING, default="by_task_type",
    help="""
    name of grouper in Drypipe.create_pipeline(...,task_groupers={'my_custom_grouper': group_func})
    """
)
def mon(pipeline, instance_dir, group_by):

    _configure_logging(None)

    from dry_pipe.cli_screen import CliScreen

    if instance_dir is None:
        instance_dir = os.getcwd()

    if pipeline is None:
        pipeline = PipelineInstance.guess_pipeline_from_hints(instance_dir)
        if pipeline is None:
            click.echo("must specify pipeline, ex: -p <aModule>:<aPipelineFunction>")

    #pipeline_instance = _pipeline_instance_creater(instance_dir, pipeline)()

    def _quit_listener(cli_screen):
        logger.debug("quit listener")
        logging.shutdown()
        cli_screen.cleanup_tty()
        os._exit(0)

    cli_screen = CliScreen(
        instance_dir,
        None,
        is_monitor_mode=True,
        quit_listener=_quit_listener,
        group_by=group_by
    )

    cli_screen.start()
    def _sigint(p1, p2):
        logger.info("received SIGINT")
        cli_screen.request_quit()

    signal.signal(signal.SIGINT, _sigint)
    signal.pause()

    logging.shutdown()
    exit(0)


@click.command()
@click.option('-p', '--pipeline', help="a_module:a_func, a function that returns a pipeline.")
@click.option('--instance-dir', type=click.Path(), default=None)
@click.option('--web-mon', is_flag=True)
@click.option('--port', default=5000, help="port for --web-mon")
@click.option('--bind', default="0.0.0.0", help="bind address for --web-mon")
@click.option('--clean', is_flag=True)
@click.option('--single', type=click.STRING, default=None,
              help="launches a single task, specified by the task key, in foreground, ex.: launch --single=TASK_KEY")
@click.option('--restart', is_flag=True, help="restart failed or killed tasks")
@click.option('--restart-failed', is_flag=True, help="restart failed tasks")
@click.option('--restart-killed', is_flag=True, help="restart killed tasks")
@click.option('--reset-failed', is_flag=True)
@click.option('--no-confirm', is_flag=True)
@click.option('--logging-conf', type=click.Path(), default=None, help="log configuration file (json)")
@click.option('--env', type=click.STRING, default=None)
@click.option(
    '--group-by', type=click.STRING, default="by_task_type",
    help="""
    name of grouper in Drypipe.create_pipeline(...,task_groupers={'my_custom_grouper': group_func})
    """
)
@click.option(
    '--regex-grouper', type=click.STRING, default=None,
    help="""
    <regex>,<format>
    example: --regex-grouper=.*\_(\w)\.(\d),group-{0}-{1}
    """
)
def run(
    pipeline, instance_dir, web_mon, port, bind,
    clean, single, restart, restart_failed, restart_killed, reset_failed,
    no_confirm, logging_conf, env, regex_grouper, group_by
):

    _configure_logging(logging_conf)

    pipeline_mod_func = pipeline

    if instance_dir is None:
        instance_dir = os.getcwd()

    if not os.path.isabs(instance_dir):
        instance_dir = os.path.abspath(os.path.expanduser(os.path.expandvars(instance_dir)))

    if PipelineState.from_pipeline_instance_dir(instance_dir, none_if_not_exists=True) is None:
        if not no_confirm:
            if not click.confirm(f"no pipeline instance exists in {instance_dir}, do you want to create one ?"):
                click.echo("no pipeline instance created, exiting.")
                return

    env_vars = None
    if env is not None:
        env_vars = {}
        for k_v in env.split(","):
            k, v = k_v.split("=")
            env_vars[k] = v

    pipeline_instance = _pipeline_instance_creater(instance_dir, pipeline_mod_func, env_vars)()

    if clean:
        pipeline_instance.clean()

    pipeline_instance.init_work_dir()

    if single is not None:
        launch_single(pipeline_instance, single)
        return

    pipeline_instance.regen_tasks_if_stale(force=True)

    for task in pipeline_instance.tasks:
        task_state = task.get_state()
        if task_state is not None:

            if (task_state.is_failed() or task_state.is_killed()) and not task.has_unsatisfied_deps():
                task.prepare()

                if restart:
                    task.re_queue()
                elif restart_failed and task_state.is_failed():
                    task.re_queue()
                elif restart_killed and task_state.is_killed():
                    task.re_queue()

    janitor = Janitor(
        pipeline_instance=pipeline_instance,
        min_sleep=0,
        max_sleep=5
    )

    janitor.start()
    janitor.start_remote_janitors()

    if web_mon:
        from dry_pipe.websocket_server import WebsocketServer
        from dry_pipe.server import PipelineUIServer
        server = PipelineUIServer.init_single_pipeline_instance(pipeline_instance.get_state())
        server.start()

        WebsocketServer.start(bind, port, instances_dir_if_has_local_janitor=pipeline_instance.pipeline_instance_dir)
        # Stopping threads break CTRL+C, let the threads die with the process exit
        #server.stop()
        #janitor.do_shutdown()
    else:
        from dry_pipe.cli_screen import CliScreen

        def _quit_listener(cli_screen):
            logger.debug("quit listener")
            logging.shutdown()
            cli_screen.cleanup_tty()
            os._exit(0)

        cli_screen = CliScreen(
            pipeline_instance.pipeline_instance_dir,
            None if regex_grouper is None else _display_grouper_func_from_spec(regex_grouper),
            is_monitor_mode=False,
            quit_listener=_quit_listener,
            group_by=group_by
        )

        cli_screen.start()
        def _sigint(p1, p2):
            logger.info("received SIGINT")

            if janitor is not None:
                janitor.request_shutdown()
            cli_screen.request_quit()

        signal.signal(signal.SIGINT, _sigint)
        signal.pause()

        logging.shutdown()
        exit(0)



    #os._exit(0)


def _display_grouper_func_from_spec(display_grouper):
    reg_fmt = display_grouper.split(",")
    if len(reg_fmt) != 2:
        raise Exception(
            f"bad format for arg --display-grouper, expects <regex>,<format> got {display_grouper}"
        )
    reg, fmt = reg_fmt
    try:
        r = re.compile(reg)
    except Exception as ex:
        raise Exception(f"--display-grouper has bad regex: {reg} ")

    fmt_insert_count = fmt.count("{")

    if r.groups != fmt_insert_count:
        raise Exception(
            f"regex match group and placeholders in format must match, got: {r.groups} != {fmt_insert_count}"
        )

    def _regex_display_grouper(task_key):
        m = r.search(task_key)
        if m is None:
            return "<Other Tasks>"
        else:
            gr = m.groups()
            if len(gr) != fmt_insert_count:
                return "<Other Tasks>"
            else:
                return fmt.format(*m.groups())
    return  _regex_display_grouper


def launch_single(pipeline, single):

    task = pipeline.tasks.get(single)

    if task is None:
        raise Exception(f"pipeline has no task with key '{single}'")

    task.re_queue()
    task.prepare()
    task_state = task.get_state()
    task_state.transition_to_launched(task.task_conf.create_executer(), task, wait_for_completion=True)
    task_state = task.get_state()
    task_state.transition_to_completed(task)

    print(f"Task {single} completed.")


@click.command()
@click.pass_context
@click.option('--instance-dir', type=click.Path(), default=None)
@click.option('--single', type=click.STRING, default=None,
              help="apply to single task, specified by the task key, in foreground, ex.: launch --single=TASK_KEY")
def requeue(ctx, single, instance_dir):

    if _bad_entrypoint(ctx):
        return

    pipeline_func = ctx.obj["pipeline_func"]

    pipeline = _pipeline_from_pipeline_func(pipeline_func, instance_dir)

    if single is None:
        raise Exception(f"unimplemented.")

    requeue_single(pipeline, single)


def requeue_single(pipeline, single):

    task = pipeline.tasks.get(single)

    if task is None:
        raise Exception(f"pipeline has no task with key '{single}'")

    task.re_queue()
    task.prepare()

    print(f"Task {single} requeued")

def _parse_instances_dir_to_pipelines(instances_dir_to_pipelines, env=None):

    for d_to_p in instances_dir_to_pipelines.split(","):

        instances_dir, mod_func = d_to_p.split("=")

        if not os.path.exists(instances_dir):
            pathlib.Path(instances_dir).mkdir(parents=True)

        pipeline = Pipeline.load_from_module_func(mod_func)
        pipeline.env_vars = env
        yield instances_dir, pipeline, mod_func

def _version():

    frame = inspect.currentframe()
    f_back = frame.f_back if frame is not None else None
    f_globals = f_back.f_globals if f_back is not None else None
    # break reference cycle
    # https://docs.python.org/3/library/inspect.html#the-interpreter-stack
    del frame

    package_name = f_globals.get("__name__")

    if package_name == "__main__":
        package_name = f_globals.get("__package__")

    if package_name:
        package_name = package_name.partition(".")[0]

    try:
        from importlib import metadata
    except ImportError:
        # Python < 3.8
        import importlib_metadata as metadata
    try:
        return metadata.version(package_name)
    except metadata.PackageNotFoundError:
        raise RuntimeError("Could not get version")


@click.command()
@click.pass_context
@click.option(
    '--instances-dir-to-pipelines',
    type=click.STRING,
    help="see 'watch' command"
)
@click.option(
    '-o',
    type=click.STRING,
    help="output file"
)
def dump_remote_task_confs(ctx, instances_dir_to_pipelines, o):

    def g():
        for instances_dir, pipeline, mod_func in _parse_instances_dir_to_pipelines(instances_dir_to_pipelines):
            yield {
                "constructor_func": mod_func,
                "pipeline_instances_dir": instances_dir,
                "remote_task_confs": [
                    tc.as_json() for tc in pipeline.remote_task_confs
                ]
            }

    if False:
        res = json.dumps({
            "dry_pipe_version": _version(),
            "watched_pipelines_config": list(g())
        }, indent = 2)
    else:
        res = list(g())

    if o is None:
        print(res)
    else:
        with open(o, "w") as o_f:
            o_f.write(res)


@click.command()
@click.pass_context
@click.option(
    '--instances-dir-to-pipelines',
    type=click.STRING,
    required=True,
    help="""
    dir1=module1:func1,...,dirN=moduleN:funcN    
      + moduleX:funcX is a function that returns a dry_pipe.Pipeline    
      + dirX is the parent directory where all pipeline_instance_dirs of PipelineX will reside  
    """
)
@click.option('--env', type=click.STRING, default=None)
def watch(ctx, instances_dir_to_pipelines, env):

    _configure_logging(None)

    janitor = Janitor(
        pipeline_instances_iterator=Pipeline.pipeline_instances_iterator({
            k: v
            for k, v, _ in _parse_instances_dir_to_pipelines(instances_dir_to_pipelines, env)
        })
    )

    janitor.start()

    janitor.start_remote_janitors()

    def _sigterm(p1, p2):
        logger.info("received SIGTERM")
        janitor.request_shutdown()
        logging.shutdown()
        time.sleep(3)
        os._exit(0)

    signal.signal(signal.SIGTERM, _sigterm)
    signal.pause()






def _validate_task_generator_callable_with_single_dsl_arg_and_get_module_file(task_generator):

    a = inspect.signature(task_generator)

    f = os.path.dirname(os.path.abspath(inspect.getmodule(task_generator).__file__))

    if len(a.parameters) != 1:
        raise Exception(f"function given to cli.run() should take a singe DryPipe.dsl() as argument, and yield tasks" +
                        f" was given {len(a.parameters)} args")

    return f


@click.command()
@click.pass_context
@click.option(
    '--instances-dir-to-pipelines',
    type=click.STRING,
    required=True,
    help="""
    dir1=module1:func1,...,dirN=moduleN:funcN    
      + moduleX:funcX is a function that returns a dry_pipe.Pipeline    
      + dirX is the parent directory where all pipeline_instance_dirs of PipelineX will reside  
    """
)
@click.option('--bind', default="0.0.0.0", help="bind address for --web-mon")
@click.option('--port', default=5000)
def serve_ui(ctx, instances_dir_to_pipelines, bind, port):
    from dry_pipe.websocket_server import WebsocketServer

    _configure_logging(None)

    def g():
        for t in instances_dir_to_pipelines.split(","):

            instances_dir, mod_func = t.split("=")

            if not os.path.exists(instances_dir):
                pathlib.Path(instances_dir).mkdir(parents=True)

            pipeline = Pipeline.load_from_module_func(mod_func)
            yield instances_dir, pipeline

    WebsocketServer.start(bind, port, dict(g()))

    os._exit(0)


def _func_from_mod_func(mod_func):

    mod, func_name = mod_func.split(":")

    if not mod.startswith("."):
        module = importlib.import_module(mod)
    else:
        module = importlib.import_module(mod[1:])

    python_task = getattr(module, func_name, None)
    if python_task is None:
        raise Exception(f"function {func_name} not found in module {mod}")

    return python_task


@click.command()
@click.pass_context
@click.argument('test-case-mod-func', type=click.STRING)
def test(ctx, test_case_mod_func):

    idx = None

    if test_case_mod_func.count(":") == 1:
        mod, func_name = test_case_mod_func.split(":")
    elif test_case_mod_func.count(":") == 2:
        mod, func_name, idx = test_case_mod_func.split(":")
        idx = int(idx)
    else:
        raise Exception(
            f"expected an argument of the form <module>:<function_name>, or <module>:<function_name>:<number>"
        )

    module = importlib.import_module(mod)

    python_call = getattr(module, func_name, None)

    if python_call is None:
        raise Exception(f"function {func_name} not found in module {mod}")

    if not isinstance(python_call, PythonCall):
        raise Exception(f"function {func_name} should be decorated with @DryPipe.python_call(tests=[...])")

    if len(python_call.tests) == 0:
        click.echo(f"python_task {test_case_mod_func} has no tests")
        return

    def run_test(python_test_case):
        init_func = python_test_case.get("init_func")

        if init_func is not None:
            init_func()

        res = python_call.func(*python_test_case["args"], test=python_test_case)

        expects = python_test_case.get("expects")
        if expects is not None:
            if res != expects:
                raise Exception(f"expected {json.dumps(expects, indent=2)}\ngot:\n{json.dumps(res, indent=2)}")

    c = 0

    if idx is None:
        for python_test_case in python_call.tests:
            run_test(python_test_case)
            print(f"test passed: {mod}:{func_name}:{c}")
            c += 1
    else:
        run_test(python_call.tests[idx])
        print(f"test passed: {test_case_mod_func}")


@click.command()
@click.pass_context
@click.argument('mod-func', type=click.STRING)
@click.option('--task-env', type=click.Path(exists=True, dir_okay=False))
def call(ctx, mod_func, task_env):

    python_task = _func_from_mod_func(mod_func)

    env = os.environ

    if task_env is not None:
        # or if pipeline provided --task=key
        env = env_from_sourcing(task_env)

    var_type_dict = {}

    for producing_task_key, var_metas, file_metas in parse_in_out_meta({
        k: v
        for k, v in env.items()
        if k.startswith("__meta_")
    }):
        #if producing_task_key != "":
        #    continue

        for var_meta in var_metas:
            name_in_producing_task, var_name, typez = var_meta
            var_type_dict[var_name] = typez

    def get_and_parse_arg(k):
        v = env.get(k)
        if v is None:
            tk = os.environ['__task_key']
            raise Exception(
                f"Task {tk} called {mod_func} with None assigned to arg {k}\n" +
                f"make sure task has var {k} declared in it's consumes clause. Ex:\n" +
                f"  dsl.task(key={tk}).consumes({k}=...)"
            )
        typez = var_type_dict.get(k)
        if typez == "int":
            return int(v)
        elif typez == "str":
            return v
        elif typez == "float":
            return float(v)
        elif typez == "glob_expression":

            class GlobExpression:
                def __call__(self, *args, **kwargs):
                    return glob.glob(os.path.expandvars(v))

                def __str__(self):
                    return os.path.expandvars(v)

            return GlobExpression()

        return v

    args = [
        get_and_parse_arg(k)
        for k, v in python_task.signature.parameters.items()
    ]

    control_dir = env["__control_dir"]

    task_logger = create_task_logger(control_dir)

    task_logger.info("will invoke PythonCall: %s(%s)", mod_func, ', '.join([str(a) for a in args]))

    out_vars = None

    try:
        out_vars = python_task.func(* args)
    except Exception as ex:
        traceback.print_exc()
        logging.shutdown()
        exit(1)

    if out_vars is not None and not isinstance(out_vars, dict):
        raise Exception(
            f"function {python_task.mod_func()} called by task {os.environ.get('__task_key')} {type(out_vars)}" +
            f"@DryPipe.python_call() can only return a python dict, or None"
        )

    try:
        if out_vars is not None:
            prev_out_vars = dict(iterate_out_vars_from(os.environ["__output_var_file"]))

            for k, v in out_vars.items():
                try:
                    v = json.dumps(v)
                    prev_out_vars[k] = v
                except TypeError as ex0:
                    print(
                        f"task call {control_dir}.{mod_func} returned var {k} of unsupported type: {type(v)}" +
                        " only primitive types are supported.",
                        file=sys.stderr
                    )
                    logging.shutdown()
                    exit(1)

            write_out_vars(prev_out_vars)

        for pid_file in glob.glob(os.path.join(control_dir, "*.pid")):
            os.remove(pid_file)

    except Exception as ex:
        traceback.print_exc()
        logging.shutdown()
        exit(1)

    logging.shutdown()


@click.command()
@click.pass_context
@click.option('--downstream-of', type=click.STRING)
@click.option('--not-matching', type=click.STRING)
@click.option('--matching', type=click.STRING)
@click.option('--instance-dir', type=click.Path(), default=None)
def reset(ctx, downstream_of, not_matching, matching, instance_dir):

    pipeline_func = ctx.obj["pipeline_func"]

    pipeline = _pipeline_from_pipeline_func(pipeline_func, instance_dir)

    if matching is not None or not_matching is not None:

        if matching is not None and not_matching is not None:
            raise Exception("--matching and --not-matching can't be both specified")

        def do_reset(key):
            if matching is not None:
                return re.match(matching, key)
            return not re.match(not_matching, key)

        for task in pipeline.tasks:
            if do_reset(task.key):
                if click.confirm(f"reset {task.key} ?"):
                    task.clean()
                    print(f"{task.key} reset")

    elif downstream_of is not None:

        after_task_including = pipeline.tasks.get(downstream_of)

        if after_task_including is None:
            raise Exception(f"pipeline has no task with key '{downstream_of}'")

        def all_upstream_deps(task):
            yield task
            for upstream_task, _1, _2 in task.upstream_deps_iterator():
                for t in all_upstream_deps(upstream_task):
                    yield t

        raise Exception("not implemented")
    else:
        if click.confirm("reset all tasks ?"):
            print("boom")

@click.command()
@click.pass_context
@click.option('--instance-dir', type=click.Path(), default=None)
def stats(ctx, instance_dir):

    the_stats = None

    if instance_dir is not None:
        the_stats = fetch_task_groups_stats(instance_dir)
    else:
        pipeline_func = ctx.obj["pipeline_func"]
        pipeline = _pipeline_from_pipeline_func(pipeline_func, instance_dir)
        the_stats = fetch_task_groups_stats(pipeline.pipeline_instance_dir)

    for stat_row in the_stats:
        print("\t".join([str(v) for v in stat_row]))


@click.command()
@click.pass_context
@click.option('-p', '--pipeline', help="a_module:a_func, a function that returns a pipeline.")
def prepare_remote_sites(ctx, pipeline):

    pipeline_spec = pipeline

    if pipeline_spec is None:
        click.echo("must specify pipeline, ex: -p <aModule>:<aPipelineFunction>")
        return

    pipeline = Pipeline.load_from_module_func(pipeline_spec)

    from rich.console import Console
    from rich.panel import Panel
    console = Console()

    for f, msg in pipeline.prepare_remote_sites_funcs():
        console.print(Panel(f"[turquoise2]{msg}"))
        with console.status("rsync upload in progress..."):
            f()

    console.print(f"done.")

@click.command()
@click.pass_context
def status(ctx):

    raise Exception(f"not implemented")

    if _bad_entrypoint(ctx):
        return

    pipeline_func = ctx.obj["pipeline_func"]

    pipeline = DryPipe.create_pipeline(pipeline_func)
    console = Console()
    console.print(_status_table(pipeline))


def _register_commands():
    cli_group.add_command(run)
    cli_group.add_command(mon)
    cli_group.add_command(prepare)
    cli_group.add_command(call)
    cli_group.add_command(test)
    cli_group.add_command(status)
    cli_group.add_command(watch)
    cli_group.add_command(serve_ui)
    cli_group.add_command(clean)
    cli_group.add_command(stats)
    cli_group.add_command(reset)
    cli_group.add_command(requeue)
    cli_group.add_command(dump_remote_task_confs)
    cli_group.add_command(prepare_remote_sites)

def run_cli():
    _register_commands()
    cli_group()

if __name__ == '__main__':
    run_cli()



import glob
import json
import logging
import os
import pathlib

PIPELINE_TRANSITIONS = {
    "running": ["stopped", "stalled", "completed"],
    "stopped": ["running"],
    "stalled": ["running"],
    "completed": ["running"]
}

PIPELINE_STATES = PIPELINE_TRANSITIONS.keys()

logger = logging.getLogger(__name__)

class PipelineState:

    @staticmethod
    def iterate_from_instances_dir(pipeline_instances_dir):
        for d in glob.glob(os.path.join(pipeline_instances_dir, "*", ".drypipe", "state.*")):
            d = os.path.dirname(d)
            yield PipelineState.from_pipeline_work_dir(d, create_if_not_exists=False)

    @staticmethod
    def from_pipeline_work_dir(work_dir, create_if_not_exists=False, none_if_not_exists=False):

        state_file = glob.glob(os.path.join(work_dir, "state.*"))

        if len(state_file) > 1:
            raise Exception(f"More than one pipeline state files: {state_file}")
        elif len(state_file) == 1:
            pipeline_state = PipelineState(state_file[0], state_file_exists=True)
        else:

            if none_if_not_exists:
                return

            state_file = [os.path.join(work_dir, "state.running")]
            state_file_exists = False
            if create_if_not_exists:
                pathlib.Path(state_file[0]).touch(exist_ok=False)
                state_file_exists = True

            pipeline_state = PipelineState(state_file[0], state_file_exists)

        return pipeline_state

    @staticmethod
    def from_pipeline_instance_dir(pipeline_instance_dir, none_if_not_exists=False):
        return PipelineState.from_pipeline_work_dir(
            os.path.join(pipeline_instance_dir, ".drypipe"),
            none_if_not_exists=none_if_not_exists
        )

    @staticmethod
    def from_state_file(state_file):
        return PipelineState(state_file, True)

    def __init__(self, state_file, state_file_exists):

        self.state_file = os.path.abspath(state_file)
        self.pipeline_instance_work_dir = os.path.dirname(self.state_file)
        self.state_file_exists = state_file_exists

        if os.path.basename(self.pipeline_instance_work_dir) != ".drypipe":
            raise Exception(f"dir {state_file} can't be a pipeline state file, must be in a dir named '.drypipe'")

        state_file_basename = os.path.basename(self.state_file)

        if not state_file_basename.startswith("state."):
            raise Exception(f"invalid statefile, should start with 'state.'")

        self.state_name = state_file_basename.split(".")[1]

        if self.state_name not in PIPELINE_STATES:
            raise Exception(f"invalid state file name {state_file}")

        self.last_update_time = self.refresh_last_update_time()

    def __str__(self):
        return f"PipelineState({self.state_name},{self.instance_dir()})"

    def refresh_last_update_time(self):
        self.last_update_time = pathlib.Path(self.pipeline_instance_work_dir).stat()
        return self.last_update_time

    def instance_dir(self):
        return os.path.dirname(self.pipeline_instance_work_dir)

    def is_completed(self):
        return self.state_name == "completed"

    def is_stalled(self):
        return self.state_name == "stalled"

    def is_running(self):
        return self.state_name == "running"

    def _transition_is_valid(self, next_state):

        valid_next_states = PIPELINE_TRANSITIONS.get(self.state_name)

        if valid_next_states is None:
            raise Exception(f"invalid pipeline state '{next_state}'")

        if next_state not in valid_next_states:
            raise Exception(f"pipeline can't transition from {self.state_name} to {next_state}")

    def touch(self):

        if len(list(glob.glob(os.path.join(self.pipeline_instance_work_dir, "state.*")))) > 0:
            return

        pathlib.Path(self.state_file).touch(exist_ok=False)
        self.state_file_exists = True

    def _transition_to(self, next_state):

        self._transition_is_valid(next_state)

        dst = os.path.join(self.pipeline_instance_work_dir, f"state.{next_state}")
        os.rename(
            self.state_file,
            dst
        )

        return dst

    def dir_basename(self):
        return os.path.basename(os.path.dirname(self.pipeline_instance_work_dir))

    def instance_dir(self):
        return os.path.dirname(self.pipeline_instance_work_dir)

    def instance_dir_key(self):

        parent, pid = self.instance_dir().split("/")[-2:]
        return f"{parent}|{pid}"

    def _counts_cache_file(self):
        return os.path.join(self.pipeline_instance_work_dir, f"counts.json")

    def transition_to_completed(self):

        from dry_pipe.monitoring import PipelineMetricsTable

        with open(self._counts_cache_file(), "w") as _f:
            _f.write(
                json.dumps(PipelineMetricsTable.totals_row(self.instance_dir()), indent=4)
            )

        self._transition_to("completed")

    def transition_to_running_if_not_running(self):
        if not self.is_running():
            self._transition_to("running")

    def totals_row(self):

        from dry_pipe.monitoring import PipelineMetricsTable

        if self.is_completed():

            cf = self._counts_cache_file()
            if os.path.exists(cf):
                with open(cf) as f:
                    return json.loads(f.read())

        return PipelineMetricsTable.totals_row(self.instance_dir())



import os
import unittest

import pipeline_with_single_bash_task
import pipeline_with_single_python_task
import pipeline_with_variable_passing
import pipeline_with_two_python_tasks
import pipeline_with_multistep_tasks_with_shared_vars

from dry_pipe import TaskConf
from dry_pipe.script_lib import env_from_sourcing
from test_utils import TestSandboxDir


class SingleTaskPipelinesTests(unittest.TestCase):

    def test_single_python_task_pipeline(self):

        d = TestSandboxDir(self)

        pipeline_instance = d.pipeline_instance_from_generator(
            pipeline_with_single_python_task.pipeline,
            completed=True
        )

        pipeline_with_single_python_task.validate_single_task_pipeline(pipeline_instance)

    def test_single_python_task_pipeline_with_container(self):

        d = TestSandboxDir(self)

        pipeline_instance = d.pipeline_instance_from_generator(
            pipeline_with_single_python_task.pipeline,
            task_conf=TaskConf(
                executer_type="process",
                container="singularity-test-container.sif"
            ),
            completed=True
        )

        pipeline_with_single_python_task.validate_single_task_pipeline(pipeline_instance)

    def test_single_bash_task_pipeline(self):

        d = TestSandboxDir(self)

        pipeline_instance = d.pipeline_instance_from_generator(
            pipeline_with_single_bash_task.pipeline,
            completed=True
        )

        pipeline_with_single_python_task.validate_single_task_pipeline(pipeline_instance)

    def test_single_bash_task_pipeline_with_container(self):

        d = TestSandboxDir(self)

        pipeline_instance = d.pipeline_instance_from_generator(
            pipeline_with_single_bash_task.pipeline,
            task_conf=TaskConf(
                executer_type="process",
                container="singularity-test-container.sif"
            ),
            completed=True
        )

        pipeline_with_single_python_task.validate_single_task_pipeline(pipeline_instance)


class MinimalistPipelinesTests(unittest.TestCase):

    def test_variable_passing(self):
        d = TestSandboxDir(self)

        pipeline_instance = d.pipeline_instance_from_generator(
            pipeline_with_variable_passing.pipeline,
            completed=True
        )

        consume_and_produce_a_var = pipeline_instance.tasks["consume_and_produce_a_var"]

        task_env_script = os.path.join(
            pipeline_instance.work_dir,
            "consume_and_produce_a_var",
            "task-env.sh"
        )

        env = env_from_sourcing(task_env_script)

        v = env.get("v")
        if v is None:
            raise Exception(f"script {task_env_script} did not resolve variable 'v'")
        v = int(v)
        self.assertEqual(v, 1234)

        self.assertEqual(consume_and_produce_a_var.out.result.fetch(), 2468)

    def test_consume_var_local_is_upstream_name(self):
        d = TestSandboxDir(self)

        pipeline_instance = d.pipeline_instance_from_generator(
            pipeline_with_two_python_tasks.pipeline,
            completed=True
        )

        pipeline_with_two_python_tasks.validate_two_task_pipeline(pipeline_instance)

    def test_variable_sharing_between_task_calls(self):

        d = TestSandboxDir(self)

        pipeline_instance = d.pipeline_instance_from_generator(
            pipeline_with_multistep_tasks_with_shared_vars.dag_generator,
            completed=True
        )

        pipeline_with_multistep_tasks_with_shared_vars.validate(self, pipeline_instance)



import importlib
import inspect
import logging
import os
import pathlib
import shutil
from itertools import groupby

from dry_pipe import TaskConf, DryPipeDsl, TaskBuilder, script_lib, bash_shebang
from dry_pipe.internals import ValidationError, SubPipeline
from dry_pipe.janitors import Janitor
from dry_pipe.pipeline_state import PipelineState
from dry_pipe.script_lib import write_pipeline_lib_script, PortablePopen
from dry_pipe.task import Task

logger = logging.getLogger(__name__)


class TaskSet:

    def __init__(self, tasks_by_keys, task_states_signature_func):
        self._tasks_by_keys = tasks_by_keys
        self._task_states_signature_func = task_states_signature_func
        self._task_states_signature = self._task_states_signature_func()

    def __iter__(self):
        for t in self._tasks_by_keys.values():
            yield t

    def __len__(self):
        return len(self._tasks_by_keys)

    def __getitem__(self, task_key):

        task = self._tasks_by_keys.get(task_key)

        if task is not None:
            return task

        raise KeyError(f"pipeline has no task {task_key}")

    def get(self, task_key):
        return self._tasks_by_keys.get(task_key)

    def is_stale(self):
        return self._task_states_signature_func() != self._task_states_signature


class Pipeline:

    @staticmethod
    def load_from_module_func(module_func_pipeline):
        """
        :param module_func_pipeline:
            <a_module_path>:<function_name>
            a_python_module.python_file:function_returns_a_pipeine
            a_module_path must be in PYTHONPATH
        :return an instance of dry_pipe.Pipeline:
        """

        mod, func_name = module_func_pipeline.split(":")

        module = importlib.import_module(mod)

        func = getattr(module, func_name, None)

        if func is None:
            raise Exception(f"function {func_name} not found in module {mod}")

        pipeline = func()

        if not isinstance(pipeline, Pipeline):
            raise Exception(f"function {func_name} in {mod} should return a Pipeline, not {type(pipeline).__name__}")

        return pipeline

    def __init__(
            self,
            generator_of_tasks,
            pipeline_code_dir=None,
            task_conf=None,
            containers_dir=None,
            env_vars=None,
            remote_task_confs=None,
            task_groupers=None,
            pipeline_code_dir_ls_command=None
    ):
        if pipeline_code_dir is None:
            pipeline_code_dir = os.path.dirname(os.path.abspath(inspect.getmodule(generator_of_tasks).__file__))

        if containers_dir is None:
            containers_dir = os.path.join(pipeline_code_dir, "containers")

        if task_conf is None:
            task_conf = TaskConf("process")

        if remote_task_confs is not None:
            for tc in remote_task_confs:
                if not isinstance(tc, TaskConf):
                    raise Exception(f"invalid type {type(tc)} given to remote_task_confs")

        def gen_task_set(pipeline_instance):

            try:
                iter(generator_of_tasks(None))
            except TypeError:
                raise ValidationError(f"function {generator_of_tasks} must return an iterator of Task (ex, via yield)")

            task_by_keys = {}

            def class_name(klass):
                return klass.__module__ + '.' + klass.__qualname__

            def _populate_task_by_keys_with_task_generator(dsl, g):

                for t_i in g(dsl):
                    if isinstance(t_i, Task):
                        # all tasks point to root pipeline_instance, even tasks of sub pipelines
                        t_i.pipeline_instance = pipeline_instance
                        task_by_keys[t_i.key] = t_i
                    elif isinstance(t_i, SubPipeline):
                        dsl2 = DryPipeDsl(
                            task_by_keys,
                            task_conf=task_conf,
                            pipeline_instance=pipeline_instance,
                            task_namespance_prefix=t_i.task_namespance_prefix
                        )
                        _populate_task_by_keys_with_task_generator(dsl2, t_i.pipeline.generator_of_tasks)
                    elif isinstance(t_i, TaskBuilder):
                        raise ValidationError(
                            f" task(key='{t_i.key}') declaration is incomplete, " +
                            "you probably need to invoke .calls(...), or '.calls(...)()' on it.")
                    else:
                        raise ValidationError(
                            f"iterator has yielded an invalid type: {type(t_i)}: '{t_i}', " +
                            f"valid types are {class_name(Task)} and {class_name(SubPipeline)}"
                        )

            _populate_task_by_keys_with_task_generator(
                DryPipeDsl(task_by_keys, task_conf=task_conf, pipeline_instance=pipeline_instance),
                generator_of_tasks
            )

            tasks = task_by_keys.values()

            def validate():
                dup_keys = [
                    k for k, task_group in groupby(tasks, lambda t: t.key) if len(list(task_group)) > 1
                ]

                if len(tasks) == 0:
                    f = os.path.abspath(inspect.getmodule(generator_of_tasks).__file__)
                    msg = f"pipeline {generator_of_tasks.__name__} defined in {f} yielded zero tasks"
                    raise Exception(msg)

                if len(dup_keys) > 0:
                    raise ValidationError(f"duplicate keys in definition: {dup_keys} value of task(key=) must be unique")

                def all_produced_files():
                    for task in tasks:
                        for file in task.all_produced_files():
                            yield task, file

                def z(task, file):
                    return file.absolute_path(task)

                for path, task_group in groupby(all_produced_files(), lambda t: z(*t)):
                    task_group = list(task_group)
                    if len(task_group) > 1:
                        task_group = ",".join(list(map(lambda t: str(t[0]), task_group)))
                        raise ValidationError(f"Tasks {task_group} have colliding output to file {path}."+
                                              " All files specified in Task(produces=) must be distinct")

            validate()

            return TaskSet(
                task_by_keys,
                lambda: [
                    task.has_completed() for task in tasks
                    # optimization: we only need to trask the states of DAG changing tasks
                    if task.key in pipeline_instance.dag_determining_tasks_ids
                ]
            )

        self.task_set_generator = gen_task_set

        self.pipeline_code_dir = pipeline_code_dir
        self.task_conf = task_conf
        self.containers_dir = containers_dir
        self.generator_of_tasks = generator_of_tasks
        self.env_vars = env_vars
        self.remote_task_confs = remote_task_confs
        self.pipeline_code_dir_ls_command = pipeline_code_dir_ls_command

        def _wrap_task_grouper(grouper_name, grouper_func):
            #if task_key.count(".") != 1:
            #    raise Exception(f"task key {task_key} violates naming convention: <prefix>.<suffix>")

            def g(task_key):
                try:
                    return grouper_func(task_key)
                except Exception as ex:
                    raise Exception(f"task grouper {grouper_name} failed on task_key: {task_key}\n{ex}")
            return g

        self.task_groupers = {
            ** {
                n: _wrap_task_grouper(n, g)
                for n, g in task_groupers.items()
            },
            "by_task_type": Task.key_grouper
        }


    def create_pipeline_instance(self, pipeline_instance_dir=None, task_conf=None, containers_dir=None, env_vars=None):

        if task_conf is not None or containers_dir is not None or env_vars is not None:
            p = Pipeline(
                generator_of_tasks=self.generator_of_tasks,
                pipeline_code_dir=self.pipeline_code_dir,
                task_conf=task_conf or self.task_conf,
                containers_dir=containers_dir or self.containers_dir,
                env_vars=env_vars or self.env_vars,
                task_groupers=self.task_groupers,
                remote_task_confs=self.remote_task_confs
            )
        else:
            p = self

        return PipelineInstance(p, pipeline_instance_dir or p.pipeline_code_dir)

    @staticmethod
    def pipeline_instances_iterator(instances_dir_for_pipeline_dict, ignore_completed=True):

        if instances_dir_for_pipeline_dict is None:
            raise Exception(f"pipeline_instances_dir can't be None")

        if len(instances_dir_for_pipeline_dict) == 0:
            raise Exception(f"pipeline_instances_dir can't be empty")

        return PipelineInstancesIterator(instances_dir_for_pipeline_dict, ignore_completed)

    def prepare_remote_sites_funcs(self):
        if self.remote_task_confs is None:
            raise Exception(f"no remote_task_confs have been assigned for this pipeline")

        for task_conf in self.remote_task_confs:

            def go():
                ssh_executer = task_conf.create_executer()
                ssh_executer.rsync_remote_code_dir_if_applies(self, task_conf)

            ssh_executer = task_conf.create_executer()
            yield go, f"Will rsync {self.pipeline_code_dir} to {ssh_executer.user_at_host()}:{task_conf.remote_pipeline_code_dir}"


        for task_conf in self.remote_task_confs:
            if task_conf.container is not None:
                ssh_executer = task_conf.create_executer()
                def go():
                    ssh_executer.rsync_remote_container(task_conf)

                yield go, f"Will upload container {task_conf.container} to "+\
                          f"{ssh_executer.user_at_host()}:{task_conf.remote_containers_dir}"

    def prepare_remote_sites(self):
        for f, msg in self.prepare_remote_sites_funcs():
            f()

class PipelineInstancesIterator:

    def __init__(self, instances_dir_for_pipeline_dict, ignore_completed):
        self.pipelines_cache = {}
        self.ignore_completed = ignore_completed
        self.instances_dir_for_pipeline_dict = instances_dir_for_pipeline_dict

    def create_instance_in(self, pipeline_instances_dir, subdir):

        pid = pathlib.Path(os.path.join(pipeline_instances_dir, subdir, ".drypipe"))
        pathlib.Path(pid).mkdir(parents=True, exist_ok=False)
        pathlib.Path(os.path.join(pid, "state.running")).touch()

    def __iter__(self):

        pipelines_by_dir = {**self.pipelines_cache}

        def gen_():

            for pipeline_instances_dir, pipeline in self.instances_dir_for_pipeline_dict.items():

                logger.debug("will iterate on pipeline instances in %s ", pipeline_instances_dir)

                for pipeline_state in PipelineState.iterate_from_instances_dir(pipeline_instances_dir):

                    if pipeline_state.is_completed() and self.ignore_completed:
                        logger.debug("pipeline completed: %s", pipeline_state.state_file)
                        continue
                    else:
                        logger.debug("pipeline NOT completed: %s, %s", pipeline_state.state_file,
                                     pipeline_state.state_name)

                    p = pipelines_by_dir.get(pipeline_state.dir_basename())

                    if p is None:
                        p = pipeline.create_pipeline_instance(pipeline_state.instance_dir())

                    yield pipeline_state.dir_basename(), p

        self.pipelines_cache = dict(gen_())

        for k, pipeline in self.pipelines_cache.items():
            yield pipeline


class PipelineInstance:

    def __init__(self, pipeline, pipeline_instance_dir):
        self.pipeline = pipeline
        self.pipeline_instance_dir = pipeline_instance_dir
        self.work_dir = os.path.join(pipeline_instance_dir, ".drypipe")
        self._publish_dir = os.path.join(pipeline_instance_dir, "output")
        self.dag_determining_tasks_ids = set()
        self.tasks = self.pipeline.task_set_generator(self)
        self._is_remote_instance_directory_prepared = set()
        self._preparation_completed = 0

    def set_remote_instance_directory_prepared(self, server_key):
        self._is_remote_instance_directory_prepared.add(server_key)

    def is_remote_instance_directory_prepared(self, server_key):
        return server_key in self._is_remote_instance_directory_prepared

    def is_preparation_completed(self):

        if self._preparation_completed >= 2:
            return True

        def is_task_preparation_completed():
            for task in self.tasks:
                task_state = task.get_state()
                if task_state is None:
                    return False
                if task.key in self.dag_determining_tasks_ids and not task_state.is_completed():
                    return False
            return True

        if is_task_preparation_completed():
            self._preparation_completed += 1

        return self._preparation_completed >= 2

    @staticmethod
    def _hint_file(instance_dir):
        return os.path.join(instance_dir, ".drypipe", "hints")

    @staticmethod
    def guess_pipeline_from_hints(instance_dir):
        return PipelineInstance.load_hints(instance_dir).get("pipeline")

    @staticmethod
    def write_hint_file_if_not_exists(instance_dir, module_func_pipeline):
        hint_file = PipelineInstance._hint_file(instance_dir)

        if not os.path.exists(hint_file):
            pathlib.Path(os.path.dirname(hint_file)).mkdir(exist_ok=True, parents=True)
            with open(hint_file, "w") as _f:
                _f.write(f"pipeline={module_func_pipeline}\n")

    @staticmethod
    def load_hints(instance_dir):
        hf = PipelineInstance._hint_file(instance_dir)
        if not os.path.exists(hf):
            return {}

        def hints():
            with open(hf) as _hf:
                for line in _hf.readlines():
                    line = line.strip()
                    if line != "":
                        k, v = line.split("=")
                        yield k.strip(), v.strip()

        return dict(hints())

    def hints(self):
        return PipelineInstance.load_hints(self.pipeline_instance_dir)

    def init_work_dir(self):

        pathlib.Path(self._publish_dir).mkdir(parents=True, exist_ok=True)
        pathlib.Path(self.work_dir).mkdir(parents=True, exist_ok=True)

        pipeline_env = os.path.join(self.work_dir, "pipeline-env.sh")
        with open(pipeline_env, "w") as f:
            f.write(f"{bash_shebang()}\n\n")
            f.write(f"export __pipeline_code_dir={self.pipeline.pipeline_code_dir}\n")
            f.write(f"export __containers_dir={self.pipeline.containers_dir}\n")
            if self.pipeline.env_vars is not None:
                for k, v in self.pipeline.env_vars.items():
                    f.write(f"export {k}={v}\n")

        os.chmod(pipeline_env, 0o764)

        #drypipe_cmds = os.path.join(self._work_dir, "dryfuncs")
        #with open(drypipe_cmds, "w") as _drypipe_cmds:
        #    _drypipe_cmds.write("#!/usr/bin/env python3\n\n")

        #    with open(os.path.join(os.path.dirname(__file__), "script_commands.py")) as f:
        #        _drypipe_cmds.write(f.read())

        #os.chmod(drypipe_cmds, 0o764)

        #drypipe_bash_lib = os.path.join(self._work_dir, "drypipe-bash-lib.sh")

        #with open(drypipe_bash_lib, "w") as f:
        #    f.write(f"{bash_shebang()}\n\n")
        #    f.write(BASH_TASK_FUNCS_AND_TRAPS)
        #    f.write(BASH_SIGN_FILES)

        #os.chmod(drypipe_bash_lib, 0o764)
        shutil.copy(script_lib.__file__, self.work_dir)

        script_lib_file = os.path.join(self.work_dir, "script_lib")
        with open(script_lib_file, "w") as script_lib_file_handle:
            write_pipeline_lib_script(script_lib_file_handle)
        os.chmod(script_lib_file, 0o764)

        #self.calc_pre_existing_files_signatures()
        self.get_state().touch()

    def remote_sites_task_confs(self, for_zombie_detection=False):

        def f(task):
            if task.is_remote():
                if not for_zombie_detection:
                    return True
                task_state = task.get_state()
                return task_state.is_step_started() or task_state.is_launched() or task_state.is_scheduled()
            else:
                return False

        return {
            task.task_conf.remote_site_key: task.task_conf
            for task in self.tasks
            if f(task)
        }.values()

    def regen_tasks_if_stale(self, force=False):
        if self.tasks.is_stale() or force:
            self.tasks = self.pipeline.task_set_generator(self)

    def instance_dir_base_name(self):
        return os.path.basename(self.pipeline_instance_dir)

    def _recalc_hash_script(self):
        return os.path.join(self.work_dir, "recalc-output-file-hashes.sh")

    def work_dir_exists(self):
        return os.path.exists(self.work_dir)

    def tasks_for_glob_expr(self, glob_expr):
        return [
            task
            for task in self.tasks
            if pathlib.PurePath(task.key).match(glob_expr)
        ]

    def tasks_for_key_prefix(self, key_prefix):
        for task in self.tasks:
            if task.key.startswith(key_prefix):
                yield task

    def clean(self):

        for task in self.tasks:
            task.clean()

    def clean_all(self):
        if os.path.exists(self.work_dir):
            shutil.rmtree(self.work_dir)

        if os.path.exists(self._publish_dir):
            shutil.rmtree(self._publish_dir)

    def get_state(self, create_if_not_exists=False):
        return PipelineState.from_pipeline_work_dir(self.work_dir, create_if_not_exists)

    def pre_existing_file_deps(self):
        def gen_deps():
            for task in self.tasks:
                for name, file in task.pre_existing_files.items():
                    yield file.file_path

        return list(gen_deps())

    def summarized_dependency_graph(self):

        def gen_deps():
            for task in self.tasks:
                for upstream_task, upstream_input_files, upstream_input_vars in task.upstream_deps_iterator():
                    yield [
                        Task.key_grouper(task.key),
                        [
                            k.produced_file.var_name
                            for k in upstream_input_files
                        ] + [
                            k.output_var.name
                            for k in upstream_input_vars
                        ],
                        Task.key_grouper(upstream_task.key)
                    ]

                for k, m in task.task_matchers.items():
                    for upstream_task in self.tasks:
                        if pathlib.PurePath(upstream_task.key).match(m.task_keys_glob_pattern):
                            yield [
                                Task.key_grouper(task.key),
                                [m.task_keys_glob_pattern],
                                Task.key_grouper(upstream_task.key)
                            ]

        def g(iterable, func):
            return groupby(sorted(iterable, key=func), key=func)

        def gen_tasks():
            for _t, tasks in g(self.tasks, lambda t: Task.key_grouper(t.key)):
                yield {
                    "key_group": _t,
                    "keys": list([t.key for t in tasks])
                }

        def gen_group_deps():
            for _d, deps in g(gen_deps(), lambda t: f"{t[0]}-{t[2]}"):
                yield next(deps)

        return {
            "taskGroups": list(gen_tasks()),
            "deps": list(gen_group_deps())
        }

    # TODO: avoid name clashes when pre existing files with the same names
    def calc_pre_existing_files_signatures(self, force_recalc=False):

        def name_file_paths_tuples():
            for task in self.tasks:
                for name, pre_existing_file in task.pre_existing_files.items():
                    p = task.abs_from_pipeline_instance_dir(pre_existing_file.absolute_path(task))

                    if not os.path.exists(p):
                        raise Exception(f"{task} depends on missing pre existing file '{name}'='{p}'")

                    yield p

        all_pre_existing_files = {
            f for f in name_file_paths_tuples()
        }

        sig_dir = os.path.join(self.work_dir, "in_sigs")

        if not os.path.exists(sig_dir):

            pathlib.Path(sig_dir).mkdir(parents=True, exist_ok=True)

            with PortablePopen(
                    f"{self.work_dir}/recalc-output-file-hashes.sh",
                    env={
                        "__control_dir": self.work_dir,
                        "__sig_dir": "in_sigs",
                        "__file_list_to_sign": ",".join(all_pre_existing_files)
                    }
            ) as p:
                p.wait_and_raise_if_non_zero()

    """
    Recompute output signatures (out.sig) of all tasks, and identifies tasks who's inputs has changed by 
    
    creating a filed named "in.sig.changed" in the task's control_dir.
    """

    def recompute_signatures(self, force_recalc_of_preexisting_file_signatures=True):

        self.calc_pre_existing_files_signatures(force_recalc_of_preexisting_file_signatures)

        for task in self.tasks:
            task.recompute_output_singature(self._recalc_hash_script())

        total_changed = 0

        for task in self.tasks:
            if task.has_completed():

                previously_computed_input_signature = task.input_signature()

                up_to_date_input_signature, in_sig_file_writer = task.calc_input_signature()

                if previously_computed_input_signature != up_to_date_input_signature:
                    total_changed += 1
                    in_sig_file_writer(write_changed_flag=True)
                else:
                    task.clear_input_changed_flag()

        return total_changed

    def run_sync(self, tmp_env_vars={}, fail_silently=False):

        for k, v in tmp_env_vars.items():
            os.environ[k] = v

        self.init_work_dir()

        j = Janitor(pipeline_instance=self)
        i = j.iterate_main_work(sync_mode=True, fail_silently=fail_silently)
        has_work = next(i)
        while has_work:
            has_work = next(i)

        for k, v in tmp_env_vars.items():
            del os.environ[k]


SLURM_SQUEUE_FORMAT_SPEC = "%A %L %j %l %T"


def parse_status_from_squeue_line(squeue_line):
    return squeue_line[4]


def call_squeue_for_job_id(job_id, parser=None):
    cmd = f"squeue -h -j {job_id} -o '{SLURM_SQUEUE_FORMAT_SPEC}'"

    with PortablePopen(cmd.split(" ")) as p:
        p.wait_and_raise_if_non_zero()
        out_line = p.stdout_as_string().strip()

        if len(out_line) == 0:
            return None
        elif len(out_line) > 1:
            raise Exception(f"squeue dared return more than one line when given {cmd}")

        line = map(lambda s: s.strip(), out_line[0].split(" "))

        if parser is not None:
            return parser(line)

        return line


def is_module_command_available():
    cmd = ["bash", "-c", "type -t module"]
    with PortablePopen(cmd) as p:
        p.wait_and_raise_if_non_zero()
        out = p.stdout_as_string().strip()
        return out == "function"
